"""LOGS GROK COMMANDS"""
import json
from pathlib import Path
import logging
import click
from clickclick import choice

from mldock.api.logs import parse_grok_multiline, get_all_file_objects
from mldock.api.assets import infer_filesystem_type

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")
MLDOCK_CONFIG_NAME = "mldock.json"


@click.group()
def grok():
    """
    Commands to manage and interact with logs using grok.
    """


@click.command()
@click.option("--log-path", type=str, help="a grok pattern")
@click.option("--log-file", type=str, default="logs.txt", help="file name")
@click.option("--pattern", type=str, help="a grok pattern")
def parse(log_path, log_file, pattern):
    """parse logs using a custom crok config and show"""

    file_system, log_path = infer_filesystem_type(log_path)

    logs = get_all_file_objects(log_path, log_file, file_system)

    log_runs = [Path(log).parents[0].name for log in logs]

    state = choice("Select a run", log_runs, default=None)

    for log in logs:
        if Path(log).parents[0].name == state:
            log_file_path = log
            break

    metadata = parse_grok_multiline(log_file_path, pattern, file_system)

    print(json.dumps(metadata, indent=4, sort_keys=True))


grok.add_command(parse)



"""PACKAGES COMMANDS"""
import logging
from pathlib import Path
import click

from mldock.api.packages import build_wheels

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")
MLDOCK_CONFIG_NAME = "mldock.json"


@click.group()
def packages():
    """
    Commands to manage package dependencies in mldock projects
    """


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.argument("pip_wheel_args", nargs=-1, type=click.UNPROCESSED)
def pack(project_directory, pip_wheel_args):
    """
    Command to create a mldock enabled container template
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception(
                (
                    "Path '{}' was not an mldock project. "
                    "Confirm this directory is correct, otherwise "
                    "create one.".format(project_directory)
                )
            )

        mldock_src_path = Path(project_directory, "src")
        container_project_packages_path = Path(mldock_src_path, "packages").as_posix()

        build_wheels(
            dist_dir=container_project_packages_path,
            pip_wheel_args=list(pip_wheel_args),
        )
    except Exception as exception:
        logger.error(exception)
        raise


packages.add_command(pack)



"""
    AWS Auth utilities
"""
import base64
import boto3

from mldock.platform_helpers.utils import strip_scheme


def get_aws_ecr(region: str):
    """
    Get Authentication credentials for AWS ECR using boto3

    args:
        region (str): AWS region of the registry to authenticate

    return:
        username (str): username to use in docker client auth
        password (str): password token to use in docker client auth
        registry (str): registry url to use
    """
    # this loads AWS access token and secret from env and returns an ECR client
    ecr_client = boto3.client("ecr", region_name=region)

    # get login token
    token = ecr_client.get_authorization_token()

    username, password = (
        base64.b64decode(token["authorizationData"][0]["authorizationToken"])
        .decode()
        .split(":")
    )

    registry = token["authorizationData"][0]["proxyEndpoint"]

    cloud_repository = strip_scheme(registry)
    # return docker credentials
    return username, password, registry, cloud_repository



"""PYARROW STORAGE HELPERS"""
import tempfile
from pathlib import Path
import logging
from pyarrow import fs

from mldock.platform_helpers import utils

logger = logging.getLogger("mldock")


def upload_assets(
    file_system, fs_base_path, local_path, storage_location, zip_artifacts: bool = False
):
    """Uploads logs to specified file-system"""

    if zip_artifacts:
        # pylint: disable=consider-using-with
        tmp_dir = tempfile.TemporaryDirectory()
        utils.zip_folder(
            local_path, Path(tmp_dir.name, "artifacts.zip"), rm_original=False
        )
        local_path = tmp_dir.name

    # create full artifacts base path
    artifacts_base_path = Path(fs_base_path, storage_location)

    local = fs.LocalFileSystem()

    file_selector = fs.FileSelector(local_path, recursive=True)
    for file in local.get_file_info(file_selector):
        src_path = Path(file.path)
        file_name = src_path.name
        dst_path = Path(artifacts_base_path, file_name)

        if isinstance(file_system, fs.LocalFileSystem):
            artifacts_base_path.mkdir(parents=True, exist_ok=True)
            file_system.copy_file(src_path.as_posix(), dst_path.as_posix())
        else:
            file_system.upload(src_path.as_posix(), dst_path.as_posix())

    if zip_artifacts:
        tmp_dir.cleanup()


def download_assets(file_system, fs_base_path, local_path, storage_location):
    """Uploads logs to specified file-system"""

    # create full artifacts base path
    artifacts_base_path = Path(fs_base_path, storage_location)

    if isinstance(file_system, fs.LocalFileSystem):
        file_selector = fs.FileSelector(fs_base_path, recursive=True)

        file_selector = file_system.get_file_info(file_selector)
    else:
        file_selector = file_system.glob(Path(artifacts_base_path, "**").as_posix())

    for file in file_selector:
        if isinstance(file, fs.FileInfo):
            file = file.path
        src_path = Path(file)
        file_name = src_path.name
        dst_path = Path(local_path, file_name)
        dst_path.parents[0].mkdir(parents=True, exist_ok=True)

        if isinstance(file_system, fs.LocalFileSystem):
            artifacts_base_path.mkdir(parents=True, exist_ok=True)
            file_system.copy_file(src_path.as_posix(), dst_path.as_posix())
        else:
            logger.info(src_path.as_posix())
            file_system.download(src_path.as_posix(), dst_path.as_posix())

            if Path(file).suffix == ".zip":

                utils.unzip_file(dst_path, local_path, rm_zipped=True)

            elif Path(file).suffix == ".gz":

                utils.unzip_file_from_tarfile(dst_path, local_path, rm_zipped=True)



"""Test local cli commands"""
import collections
import uuid
from pathlib import Path
import tempfile
import pytest
from mock import patch
from click.testing import CliRunner

from mldock.__main__ import cli
from mldock.platform_helpers import utils

@pytest.fixture
def log_data():

    return (
        "Experiment id = landing_page_forecaster\n"
        "param: train:ESTIMATORS=10;\n"
        "param: test#iterations=5;\n"
        "Running Evaluation\n"
        "metric: train:accuracy=0.9666666666666667;\n"
        "metric: test~accuracy=0.9;\n"
        "Refitting Final Model"
    )


class TestLocalCommands:
    """Test local cli commmands"""

    @staticmethod
    def __create_textfile(my_path, log_data):
        """creates textfile and seeds it with msg"""
        # This currently assumes the path to the file exists.
        with open(my_path, 'w+') as file:
            file.write(log_data)

    @patch('mldock.api.logs.print_table')
    def test_logs_metrics_grok_successful(self, mock_print_table, log_data):
        """test logs metrics groks logs as aspected"""
        runner = CliRunner()

        # create tmp directory
        tmp_dir = tempfile.TemporaryDirectory()

        # create tmp experiment dir
        tmp_experiment_dir = Path(tmp_dir.name, str(uuid.uuid4()))
        tmp_experiment_dir.mkdir(parents=True, exist_ok=True)

        # create textfile of logs
        txtfile = Path(tmp_experiment_dir, 'logs.txt').as_posix()
        self.__create_textfile(txtfile, log_data=log_data)

        # run CLI command
        result = runner.invoke(
            cli=cli,
            args=['logs', 'metrics', 'show', '--log-path', tmp_dir.name]
        )

        # mock print_table and get args it was passed
        args, _ = mock_print_table.call_args
        validation = {'experiment', 'test~accuracy', 'run_id', 'train:accuracy'}

        assert args[0] == validation, "Failed to parse metric names correctly"
        assert result.exit_code == 0, result.output

    @patch('mldock.api.logs.print_table')
    def test_logs_params_grok_successful(self, mock_print_table, log_data):
        """test logs params groks logs as aspected"""
        runner = CliRunner()

        # create tmp directory
        tmp_dir = tempfile.TemporaryDirectory()

        # create tmp experiment dir
        tmp_experiment_dir = Path(tmp_dir.name, str(uuid.uuid4()))
        tmp_experiment_dir.mkdir(parents=True, exist_ok=True)

        # create textfile of logs
        txtfile = Path(tmp_experiment_dir, 'logs.txt').as_posix()
        self.__create_textfile(txtfile, log_data=log_data)

        # run CLI command
        result = runner.invoke(
            cli=cli,
            args=['logs', 'params', 'show', '--log-path', tmp_dir.name]
        )

        # mock print_table and get args it was passed
        args, _ = mock_print_table.call_args
        validation = {'test#iterations', 'experiment', 'run_id', 'train:ESTIMATORS'}

        assert args[0] == validation, "Failed to parse param names correctly"
        assert result.exit_code == 0, result.output



"""Registry API methods"""
import logging
import click
import docker

logger = logging.getLogger("mldock")


def compute_progress(current, max_value=10):
    """
    compute the progress bar update

    args:
        current (int): current count of progress bar
        max_value (int): (default=10) maximum length of progress bar
    returns:
        (int): updated progress bar
    """
    if current > max_value:
        multiple = current // max_value
        current = current - max_value * multiple

    return current


def get_layer_state(stream: str, layer_id: str, state: dict = None, **kwargs):
    """
    Computes a status update to the state of layer during a push.
    Additionally, adds a progress bar and flexible fill character.

    args:
        stream (str): stream message
        layer_id (str): stream layer id
        state (dict) (default=None) current state
    kwargs:
        progress (str): (default = '') starting progress bar string
        fill_char (str): (default='=') fill char to use as progress bar
        increment (str): (default=2) progress bar char increment to use
        max_value (str): (default=40) maximum progress bar length
        status_tag (str): (default='pushing') key to use to look for updates
    """
    progress = kwargs.get("progress", "")
    fill_char = kwargs.get("fill_char", "=")
    increment = kwargs.get("increment", 2)
    max_value = kwargs.get("max_value", 40)
    status_tag = kwargs.get("status_tag", "pushing")

    try:

        if status_tag.lower() == "pushing":
            # progress = ''
            if stream.lower() == status_tag.lower():
                current_layer = state[layer_id]
                current_progress = (
                    current_layer["progress"]
                    .replace("[", "")
                    .replace("]", "")
                    .replace(">", "")
                )
                current_progress = len(current_progress)
                progress = (
                    compute_progress(current_progress, max_value=max_value) + increment
                )
                progress = fill_char * progress
                progress = "[{}>]".format(progress)

        state.update({layer_id: {"message": stream, "progress": progress}})
    except KeyError:
        pass

    return state


def stateful_log_emitter(line: dict, status_tag: str, states: dict = None):
    """
    stateful log emitter

    args:
        line (dict): current record of logs
        status_tag (str): key to use to update state
        states (dict): current state of logs
    """
    if states is None:
        states = {}

    error = line.get("error", None)
    error_detail = line.get("errorDetail", None)
    if error is not None:
        logger.error("{}\n{}".format(error, error_detail))

    stream = line.get("status", "")
    layer_id = line.get("id", "")
    progress = line.get("progress", "")

    # perform status update, using layer_id as key
    states = get_layer_state(
        stream, layer_id, states, progress=progress, status_tag=status_tag
    )

    # clear terminal to allow for friendlier update
    click.clear()

    # emit current state to stdout
    for key, value in states.items():
        logger.info(
            "{KEY}: {MESSAGE} {PROGRESS}".format(
                KEY=key, MESSAGE=value["message"], PROGRESS=value["progress"]
            )
        )

    return states


def push_image_to_repository(image_repository: str, auth_config: dict, tag="latest"):
    """
    Push image to repository in registry. Using auth_config this will Authenticate client.
    """
    client = docker.from_env()

    push_response = client.images.push(
        repository=image_repository,
        tag=tag,
        stream=True,
        decode=True,
        auth_config=auth_config,
    )

    states = {}
    for line in push_response:
        states = stateful_log_emitter(line=line, states=states, status_tag="Pushing")
    return states


def pull_image_from_repository(image_repository: str, auth_config: dict, tag="latest"):
    """
    Pull image from repository in registry. Using auth_config this will Authenticate client.
    """
    client = docker.from_env()

    pull_response = client.api.pull(
        repository=image_repository,
        tag=tag,
        stream=True,
        decode=True,
        auth_config=auth_config,
    )

    states = {}
    for line in pull_response:
        states = stateful_log_emitter(
            line=line, states=states, status_tag="Downloading"
        )
    return states



"""REGISTRY COMMANDS"""
import os
import logging
import click

from mldock.platform_helpers.docker.auth import login_and_authenticate
from mldock.api.local import docker_build
from mldock.api.registry import push_image_to_repository, pull_image_from_repository
from mldock.config_managers.project import MLDockConfigManager
from mldock.terminal import ProgressLogger, pretty_build_logs

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")
MLDOCK_CONFIG_NAME = "mldock.json"


def reset_terminal():
    """clears the terminal view frame"""
    click.clear()


@click.group()
def registry():
    """
    Commands to interact with docker image registries.
    """


@click.command()
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option("--no-cache", help="builds container from scratch", is_flag=True)
@click.option(
    "--build", help="Set the working directory for your mldock container.", is_flag=True
)
@click.option(
    "--provider",
    help="Set the cloud provider",
    required=True,
    type=click.Choice(["ecr", "gcr", "dockerhub"], case_sensitive=False),
)
@click.option("--region", help="Set the registry region", default=None, type=str)
@click.option("--tag", help="docker tag", type=str, default="latest")
@click.option("--stage", help="environment to stage.")
@click.pass_obj
def push(obj, project_directory, provider, **kwargs):
    """
    Command to push docker container image to Image Registry
    """
    tag = kwargs.get("tag", "latest")
    stage = kwargs.get("stage", None)
    region = kwargs.get("region", None)
    build = kwargs.get("build", False)
    reset_terminal()
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_dir name
    mldock_config = mldock_manager.get_config()
    module_path = os.path.join(
        project_directory,
        mldock_config.get("mldock_module_dir", "src"),
    )
    dockerfile_path = os.path.join(
        project_directory,
        mldock_config.get("mldock_module_dir", "src"),
        mldock_config.get("container_dir", None),
        "Dockerfile",
    )
    requirements_file_path = os.path.join(
        project_directory, mldock_config.get("requirements_dir", "requirements.txt")
    )

    # retrieve stages
    with ProgressLogger(
        group="Stages",
        text="Retrieving Stages",
        spinner="dots",
        on_success="Stages Retrieved",
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]["tag"]

        if tag is None:
            raise Exception(
                "tag is not valid. Either choose a stage or set a tag manually"
            )

    # athenticate and login docker client
    with ProgressLogger(
        group="Authentication",
        text="Authenticating with {}".format(provider),
        spinner="dots",
    ) as spinner:
        _, metadata = login_and_authenticate(provider=provider, region=region)
        image_repository = (
            f"{metadata['repository']}/{mldock_config.get('image_name', None)}"
        )

    if build:
        # build image for cloud repository
        with ProgressLogger(group="Build", text="Building", spinner="dots") as spinner:
            logs = docker_build(
                image_name=image_repository,
                dockerfile_path=dockerfile_path,
                module_path=module_path,
                target_dir_name=mldock_config.get("mldock_module_dir", "src"),
                requirements_file_path=requirements_file_path,
                no_cache=kwargs.get("no_cache", False),
                docker_tag=tag,
            )
            for line in logs:

                spinner.info(pretty_build_logs(line=line))

                spinner.start()

    # Push image to cloud repository
    with ProgressLogger(
        group="Push", text="Pushing to {}".format(image_repository), spinner="dots"
    ) as spinner:
        states = push_image_to_repository(
            image_repository=image_repository,
            auth_config={
                "username": metadata["username"],
                "password": metadata["password"],
            },
            tag=tag,
        )

        for layer_id, metadata in states.items():

            single_line = "{} {}".format(layer_id, metadata["message"])

            spinner.info(single_line)

        reset_terminal()
        logger.info(obj["logo"])
        spinner.start()


@click.command()
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--provider",
    help="Set the cloud provider",
    required=True,
    type=click.Choice(["ecr", "gcr", "dockerhub"], case_sensitive=False),
)
@click.option("--region", help="Set the registry region", default=None, type=str)
@click.option("--tag", help="docker tag", type=str, default="latest")
@click.option("--stage", help="environment to stage.")
@click.pass_obj
def pull(obj, project_directory, provider, **kwargs):
    """
    Command to pull docker container image from Image Registry
    """
    tag = kwargs.get("tag", "latest")
    stage = kwargs.get("stage", None)
    region = kwargs.get("region", None)
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
    )

    # get mldock_module_dir name
    mldock_config = mldock_manager.get_config()

    # retrieve stages
    with ProgressLogger(
        group="Stages",
        text="Retrieving Stages",
        spinner="dots",
        on_success="Stages Retrieved",
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]["tag"]

        if tag is None:
            raise Exception(
                "tag is not valid. Either choose a stage or set a tag manually"
            )

    # login and authenticate
    with ProgressLogger(
        group="Authentication",
        text="Authenticating with {}".format(provider),
        spinner="dots",
    ) as spinner:
        _, metadata = login_and_authenticate(provider=provider, region=region)
        image_repository = (
            f"{metadata['repository']}/{mldock_config.get('image_name', None)}"
        )

    # Push image to cloud repository
    with ProgressLogger(
        group="Pull", text="Pulling from {}".format(image_repository), spinner="dots"
    ) as spinner:
        states = pull_image_from_repository(
            image_repository=image_repository,
            auth_config={
                "username": metadata["username"],
                "password": metadata["password"],
            },
            tag=tag,
        )

        for layer_id, metadata in states.items():
            single_line = "{} {}".format(layer_id, metadata["message"])

            spinner.info(single_line)

        reset_terminal()
        logger.info(obj["logo"])
        spinner.start()


def add_commands(cli_group: click.group):
    """
    add commands to cli group
    args:
        cli (click.group)
    """
    cli_group.add_command(push)
    cli_group.add_command(pull)


add_commands(registry)



"""LOGS METRICS COMMANDS"""
import logging
import click
from mldock.api.logs import extract_records_with_pattern

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")

MLDOCK_CONFIG_NAME = "mldock.json"


@click.group()
def metrics():
    """
    Commands to manage and interact with metric logs.
    """


@click.command()
@click.option("--log-path", type=str, help="a grok pattern")
@click.option("--log-file", type=str, default="logs.txt", help="file name")
def show(log_path, log_file):
    """show metrics for all runs as a table"""

    pattern = r"metric: %{GREEDYDATA:name}=%{NUMBER:value};"

    extract_records_with_pattern(pattern, log_path, log_file)


def add_commands(cli_group: click.group):
    """
    add commands to cli group
    args:
        cli (click.group)
    """
    cli_group.add_command(show)


add_commands(metrics)



"""MODEL BASE CLASS"""
from abc import ABCMeta, abstractmethod


class ModelService:
    """
    ModelService wraps up all preprocessing, inference and postprocessing
    functions used by model service. It is defined in a flexible manner to
    be easily extended to support different frameworks.
    """

    __metaclass__ = ABCMeta

    model = None

    def __init__(self, model_path: str):
        """
        Internal initialize ModelService.

        args:
            model_path (str): path to model
        return:
            model object
        """
        self.model = self.load_model(model_path)

    @staticmethod
    @abstractmethod
    def load_model(model_path: str):
        """
        (Abstract) load model

        args:
            model_path(str): path to model
        """
        # pylint: disable=unnecessary-pass
        pass

    @staticmethod
    @abstractmethod
    def predict(input_data):
        """
        (Abstract) Predict method

        args:
            input : iterable of records model takes at prediction time.
        return:
            iterable of outputs from model
        """
        # pylint: disable=unnecessary-pass
        pass

    @staticmethod
    @abstractmethod
    def input_transform(input_data, **kwargs):
        """
        (Abstract) Output Transform

        args:
            input: raw iterable from request
        return:
            input transformed data
        """
        # pylint: disable=unnecessary-pass
        pass

    @staticmethod
    @abstractmethod
    def output_transform(predictions, **kwargs):
        """
        (Abstract) Output Transform

        args:
            predictions: predictions returned from model.predict
        return:
            output transformed predictions
        """
        # pylint: disable=unnecessary-pass
        pass



"""LOCAL COMMANDS"""
import os
import sys
import json
from pathlib import Path
import logging
import subprocess
import click

from mldock.config_managers.cli import CliConfigureManager
import mldock.api.predict as predict_request
from mldock.terminal import (
    ChoiceWithNumbers,
    style_dropdown,
    style_2_level_detail,
    ProgressLogger,
    pretty_build_logs,
)
from mldock.platform_helpers.mldock import utils as mldock_utils
from mldock.config_managers.project import MLDockConfigManager
from mldock.api.local import (
    search_mldock_containers,
    docker_build,
    train_model,
    deploy_model,
)

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")
MLDOCK_CONFIG_NAME = "mldock.json"


def run_script_interactively(cmd, **kwargs):
    """Basic runner script for local interactive script execution"""
    output = subprocess.check_output(cmd, **kwargs)
    logger.info(output)


def python_executable():
    """Return the real path for the Python executable, if it exists.
    Return RuntimeError otherwise.
    Returns:
        (str): The real path of the current Python executable.
    """
    if not sys.executable:
        raise RuntimeError(
            "Failed to retrieve the real path for the Python executable binary"
        )
    return sys.executable


@click.group()
def local():
    """
    Commands for local development
    """


@click.command()
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option("--no-cache", help="builds container from scratch", is_flag=True)
@click.option("--tag", help="docker tag", type=str, default="latest")
@click.option("--stage", help="environment to stage.")
def build(project_directory, no_cache, tag, stage):
    """Command to build container locally"""
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_dir name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)
    template_name = mldock_config.get("template", None)
    container_dir = mldock_config.get("container_dir", None)
    module_path = os.path.join(
        project_directory,
        mldock_config.get("mldock_module_dir", "src"),
    )
    dockerfile_path = os.path.join(
        project_directory,
        mldock_config.get("mldock_module_dir", "src"),
        container_dir,
        "Dockerfile",
    )
    requirements_file_path = os.path.join(
        project_directory, mldock_config.get("requirements_dir", "requirements.txt")
    )

    with ProgressLogger(
        group="Stages",
        text="Retrieving Stages",
        spinner="dots",
        on_success="Stages Retrieved",
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]["tag"]

        if tag is None:
            raise Exception(
                "tag is not valid. Either choose a stage or set a tag manually"
            )

    try:
        with ProgressLogger(group="Build", text="Building", spinner="dots") as spinner:
            logs = docker_build(
                image_name=image_name,
                dockerfile_path=dockerfile_path,
                module_path=module_path,
                target_dir_name=mldock_config.get("mldock_module_dir", "src"),
                requirements_file_path=requirements_file_path,
                no_cache=no_cache,
                docker_tag=tag,
                container_platform=template_name,
            )
            for line in logs:

                spinner.info(pretty_build_logs(line=line))

                spinner.start()

    except Exception as exception:
        logger.error(exception)
        raise


from enum import Enum


class AuthType(Enum):

    BEARER = "bearer"


@click.command()
@click.option(
    "--payload",
    help="Path to payload",
    required=True,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--request-content-type",
    default="application/json",
    help="format of payload",
    type=click.Choice(
        ["application/json", "text/csv", "image/jpeg"], case_sensitive=False
    ),
)
@click.option(
    "--response",
    help="Path to payload",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--response-content-type",
    default="application/json",
    help="format of payload",
    type=click.Choice(
        ["application/json", "text/csv", "image/jpeg"], case_sensitive=False
    ),
)
@click.option(
    "--host",
    help="host url at which model is served",
    type=str,
    default="http://127.0.0.1:8080/invocations",
)
@click.option(
    "--headers",
    help="(Optional) Authentication to use for request",
    type=click.STRING,
    multiple=True,
)
def predict(payload, host, **kwargs):
    """
    Command to execute prediction request against ml endpoint
    """
    headers = {}

    for header in kwargs.get("headers"):
        headers.update(json.loads(header))

    with ProgressLogger(group="Predict", text="Running Request", spinner="dots"):
        if payload is None:
            logger.info(
                "\nPayload cannot be None. Please provide path to payload file."
            )
        else:
            pretty_output = predict_request.handle_prediction(
                host=host,
                request=payload,
                response_file=kwargs.get("response"),
                request_content_type=kwargs.get(
                    "request_content_type", "application/json"
                ),
                response_content_type=kwargs.get(
                    "response_content_type", "application/json"
                ),
                headers=headers,
            )
            logger.info(pretty_output)


@click.command()
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--params",
    "-p",
    help="(Optional) Hyperparameter override when running container.",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
)
@click.option(
    "--env_vars",
    "-e",
    help="(Optional) Environment Variables override when running container.",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
)
@click.option("--tag", help="docker tag", type=str, default="latest")
@click.option("--stage", help="environment to stage.")
@click.option("--interactive", help="run workflow without docker", is_flag=True)
def train(project_directory, **kwargs):
    """
    Command to run training locally on localhost
    """
    params = kwargs.get("params", None)
    env_vars = kwargs.get("env_vars", None)
    tag = kwargs.get("tag", None)
    stage = kwargs.get("stage", None)
    interactive = kwargs.get("interactive", False)

    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_path name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)

    with ProgressLogger(
        group="Stages",
        text="Retrieving Stages",
        spinner="dots",
        on_success="Stages Retrieved",
    ) as spinner:
        stages = mldock_config.get("stages", "null")

        if stage is not None:
            tag = stages[stage]["tag"]

        if tag is None:
            raise Exception(
                "tag is not valid. Either choose a stage or set a tag manually"
            )

    with ProgressLogger(
        group="Environment",
        text="Retrieving Environment Varaiables",
        spinner="dots",
        on_success="Environment Ready",
    ) as spinner:
        project_env_vars = mldock_config.get("environment", {})
        for env_var in env_vars:
            key_, value_ = env_var
            project_env_vars.update({key_: value_})

        hyperparameters = mldock_config.get("hyperparameters", {})

        # override hyperparameters
        for param in params:
            key_, value_ = param
            hyperparameters.update({key_: value_})

        env_vars = mldock_utils.collect_mldock_environment_variables(
            stage=stage, hyperparameters=hyperparameters, **project_env_vars
        )

        config_manager = CliConfigureManager()
        env_vars.update(config_manager.local.get("environment", {}))

    with ProgressLogger(
        group="Training", text="Running Training", spinner="dots"
    ) as spinner:

        spinner.info("Training Environment = {}".format(env_vars))
        spinner.start()
        if interactive:
            spinner.info("Running interactively")
            spinner.start()
            # just run python src/container/prediction/serve.py
            base_ml_path = project_directory
            # must update /opt/ml working directory before running
            # perhaps setting from environment would be the best
            env_vars.update(
                {
                    "MLDOCK_BASE_DIR": Path(base_ml_path).absolute().as_posix(),
                    "MLDOCK_INPUT_DIR": ".",
                }
            )

            # subprocess only supports 'ascii' supported str for environment variables
            env_vars = mldock_utils.format_dict_for_subprocess(env_vars)
            script_path = "src/container/training/train.py"
            run_script_interactively(
                [python_executable(), script_path], cwd=base_ml_path, env=env_vars
            )
        else:
            spinner.info("Running docker container")
            spinner.start()
            train_model(
                working_dir=project_directory,
                docker_tag=tag,
                image_name=image_name,
                entrypoint="src/container/executor.sh",
                cmd="train",
                env=env_vars,
            )


@click.command()
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--params",
    "-p",
    help="(Optional) Hyperparameter override when running container.",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
)
@click.option(
    "--env_vars",
    "-e",
    help="(Optional) Environment Variables override when running container.",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
)
@click.option("--tag", help="docker tag", type=str, default="latest")
@click.option(
    "--port", help="host url at which model is served", type=str, default="8080"
)
@click.option("--stage", help="environment to stage.")
@click.option("--interactive", help="run workflow without docker", is_flag=True)
@click.pass_obj
def deploy(obj, project_directory, **kwargs):
    """
    Command to deploy ml container on localhost
    """
    params = kwargs.get("params", None)
    env_vars = kwargs.get("env_vars", None)
    tag = kwargs.get("tag", None)
    port = kwargs.get("port", None)
    stage = kwargs.get("stage", None)
    interactive = kwargs.get("interactive", False)

    verbose = obj.get("verbose", False)
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_path name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)

    with ProgressLogger(
        group="Stages",
        text="Retrieving Stages",
        spinner="dots",
        on_success="Stages Retrieved",
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]["tag"]

        if tag is None:
            raise Exception(
                "tag is not valid. Either choose a stage or set a tag manually"
            )

    project_env_vars = mldock_config.get("environment", {})

    for env_var in env_vars:
        key_, value_ = env_var
        print(key_, value_)
        project_env_vars.update({key_: value_})

    hyperparameters = mldock_config.get("hyperparameters", {})

    # override hyperparameters
    for param in params:
        key_, value_ = param
        print(key_, value_)
        hyperparameters.update({key_: value_})

    env_vars = mldock_utils.collect_mldock_environment_variables(
        stage=stage,
        hyperparameters=mldock_config.get("hyperparameters", {}),
        **project_env_vars
    )

    config_manager = CliConfigureManager()
    env_vars.update(config_manager.local.get("environment", {}))

    with ProgressLogger(
        group="Deploy",
        text="Deploying model to {} @ localhost".format(port),
        spinner="dots",
    ) as spinner:
        spinner.info("Deployment Environment = {}".format(env_vars))
        spinner.start()
        if interactive:
            spinner.info("Running interactively")
            spinner.start()
            # just run python src/container/prediction/serve.py
            base_ml_path = project_directory
            # must update /opt/ml working directory before running
            # perhaps setting from environment would be the best
            # check out the python runner from sagemaker_training
            env_vars.update(
                {
                    "MLDOCK_BASE_DIR": Path(base_ml_path).absolute().as_posix(),
                    "MLDOCK_INPUT_DIR": ".",
                }
            )

            # subprocess only supports 'ascii' supported str for environment variables
            env_vars = mldock_utils.format_dict_for_subprocess(env_vars)

            script_path = "src/container/prediction/serve.py"

            run_script_interactively(
                [python_executable(), script_path], cwd=base_ml_path, env=env_vars
            )
        else:
            spinner.info("Running docker container")
            spinner.start()

            deploy_model(
                working_dir=project_directory,
                docker_tag=tag,
                image_name=image_name,
                port={8080: port},
                entrypoint="src/container/executor.sh",
                cmd="serve",
                env=env_vars,
                verbose=verbose,
            )


@click.command()
def stop():
    """
    Command to stop ml container on localhost
    """

    try:
        containers = search_mldock_containers()

        tags = [
            style_2_level_detail(major_detail=c.image.tags[0], minor_detail=c.name)
            for c in containers
        ] + ["abort"]
        # newline break
        click.echo("")
        click.secho("Running MLDock containers:", bg="blue", nl=True)
        container_tag = click.prompt(
            text=style_dropdown(
                group_name="container name", options=tags, default="abort"
            ),
            type=ChoiceWithNumbers(tags, case_sensitive=False),
            show_default=False,
            default="abort",
            show_choices=False,
        )

        for container in containers:
            for image_tag in container.image.tags:
                if image_tag in container_tag:
                    # make newline
                    click.echo("")
                    project_container_tag = click.style(container_tag, fg="bright_blue")
                    if click.confirm(
                        "Stop the running container = {}?".format(project_container_tag)
                    ):
                        container.kill()

                    break

    except Exception as exception:
        logger.error(exception)
        raise


def add_commands(cli_group: click.group):
    """
    add commands to cli group
    args:
        cli (click.group)
    """
    cli_group.add_command(build)
    cli_group.add_command(predict)
    cli_group.add_command(train)
    cli_group.add_command(deploy)
    cli_group.add_command(stop)


add_commands(local)



"""TEMPLATES COMMANDS"""
import mimetypes
import logging
from pathlib import Path
import click

from mldock.config_managers.cli import InputDataConfigManager, CliConfigureManager
from mldock.config_managers.project import MLDockConfigManager
from mldock.terminal import ProgressLogger
from mldock.platform_helpers.mldock.storage.pyarrow import (
    upload_assets,
    download_assets,
)
from mldock.api.assets import infer_filesystem_type

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")
MLDOCK_CONFIG_NAME = "mldock.json"


@click.group()
def datasets():
    """
    Commands to create, update and manage datasets for container projects
    """


@click.command()
@click.option(
    "--channel",
    help="asset channel name. Directory name, within project data/ to store assets",
    required=True,
    type=str,
)
@click.option(
    "--name",
    help="asset filename name. File name, within project data/<channel> in which data artifact will be found",
    required=True,
    type=str,
)
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--remote", help="mldock remote to use to store or fetch dataset", type=str
)
@click.option("--remote_path", help="relative path within remote to store", type=str)
@click.option("--mime_type", "--type", help="type of file based on mimetypes", type=str)
@click.option(
    "--compression",
    help="type of file based on mimetypes",
    type=click.Choice(["zip"], case_sensitive=False),
)
def create(
    channel, name, project_directory, remote, remote_path, mime_type, compression
):
    """
    Command to create dataset manifest for mldock enabled container projects.
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception(
                (
                    "Path '{}' was not an mldock project. "
                    "Confirm this directory is correct, otherwise "
                    "create one.".format(project_directory)
                )
            )

        if mime_type is None:
            mime_type = mimetypes.guess_type(name)
            if isinstance(mime_type, (list, tuple)):
                mime_type = mime_type[0]

        if remote_path is None:
            remote_path = channel

        mldock_manager = MLDockConfigManager(
            filepath=Path(project_directory, MLDOCK_CONFIG_NAME)
        )

        # get mldock_module_dir name
        mldock_config = mldock_manager.get_config()

        input_data_channels = InputDataConfigManager(
            config=mldock_config.get("data", []),
            base_path=Path(project_directory, "data"),
        )

        dataset = input_data_channels.get(
            channel=channel,
            filename=name,
        )

        if compression is None:
            compression = dataset.get("compression", None)

        if remote is None:
            remote = dataset.get("remote", None)

        input_data_channels.add_asset(
            channel=channel,
            filename=name,
            type=mime_type,
            remote=remote,
            compression=compression,
            remote_path=remote_path,
        )
        input_data_channels.write_gitignore()
        mldock_manager.update_data_channels(data=input_data_channels.get_config())

        mldock_manager.write_file()

    except Exception as exception:
        logger.error(exception)
        raise


@click.command()
@click.option(
    "--channel",
    help="asset channel name. Directory name, within project data/ to store assets",
    required=True,
    type=str,
)
@click.option(
    "--name",
    help="asset filename name. File name, within project data/<channel> in which data artifact will be found",
    required=True,
    type=str,
)
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--remote", help="mldock remote to use to store or fetch dataset", type=str
)
@click.option("--remote_path", help="relative path within remote to store", type=str)
@click.option("--mime_type", "--type", help="type of file based on mimetypes", type=str)
@click.option(
    "--compression",
    help="type of file based on mimetypes",
    type=click.Choice(["zip"], case_sensitive=False),
)
def update(
    channel, name, project_directory, remote, remote_path, mime_type, compression
):
    """
    Command to create dataset manifest for mldock enabled container projects.
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception(
                (
                    "Path '{}' was not an mldock project. "
                    "Confirm this directory is correct, otherwise "
                    "create one.".format(project_directory)
                )
            )

        mldock_manager = MLDockConfigManager(
            filepath=Path(project_directory, MLDOCK_CONFIG_NAME)
        )

        # get mldock_module_dir name
        mldock_config = mldock_manager.get_config()

        input_data_channels = InputDataConfigManager(
            config=mldock_config.get("data", []),
            base_path=Path(project_directory, "data"),
        )

        dataset = input_data_channels.get(
            channel=channel,
            filename=name,
        )

        if mime_type is None:
            mime_type = dataset.get("type", None)

        if compression is None:
            compression = dataset.get("compression", None)
        
        if remote_path is None:
            remote_path = dataset.get("remote_path", None)

        if remote is None:
            remote = dataset.get("remote", None)

        input_data_channels.add_asset(
            channel=channel,
            filename=name,
            type=mime_type,
            remote=remote,
            compression=compression,
            remote_path=remote_path,
            update=True,
        )
        input_data_channels.write_gitignore()
        mldock_manager.update_data_channels(data=input_data_channels.get_config())

        mldock_manager.write_file()

    except Exception as exception:
        logger.error(exception)
        raise


@click.command()
@click.option(
    "--channel",
    help="asset channel name. Directory name, within project data/ to store assets",
    required=True,
    type=str,
)
@click.option(
    "--name",
    help="asset filename name. File name, within project data/<channel> in which data artifact will be found",
    required=True,
    type=str,
)
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
def remove(channel, name, project_directory):
    """
    Command to create dataset manifest for mldock enabled container projects.
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception(
                (
                    "Path '{}' was not an mldock project. "
                    "Confirm this directory is correct, otherwise "
                    "create one.".format(project_directory)
                )
            )

        mldock_manager = MLDockConfigManager(
            filepath=Path(project_directory, MLDOCK_CONFIG_NAME)
        )

        # get mldock_module_dir name
        mldock_config = mldock_manager.get_config()

        input_data_channels = InputDataConfigManager(
            config=mldock_config.get("data", []),
            base_path=Path(project_directory, "data"),
        )

        input_data_channels.remove(channel=channel, filename=name)
        input_data_channels.write_gitignore()
        mldock_manager.update_data_channels(data=input_data_channels.get_config())

        mldock_manager.write_file()

    except Exception as exception:
        logger.error(exception)
        raise


@click.command()
@click.option(
    "--channel",
    help="asset channel name. Directory name, within project data/ to store assets",
    required=True,
    type=str,
)
@click.option(
    "--name",
    help="asset filename name. File name, within project data/<channel> in which data artifact will be found",
    required=True,
    type=str,
)
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
def push(channel, name, project_directory):
    """
    Command to create dataset manifest for mldock enabled container projects.
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception(
                (
                    "Path '{}' was not an mldock project. "
                    "Confirm this directory is correct, otherwise "
                    "create one.".format(project_directory)
                )
            )

        mldock_manager = MLDockConfigManager(
            filepath=Path(project_directory, MLDOCK_CONFIG_NAME)
        )

        # get mldock_module_dir name
        mldock_config = mldock_manager.get_config()

        input_data_channels = InputDataConfigManager(
            config=mldock_config.get("data", []),
            base_path=Path(project_directory, "data"),
        )

        dataset = input_data_channels.get(channel=channel, filename=name)

        config_manager = CliConfigureManager()

        remote = config_manager.remotes.get(name=dataset["remote"])

        file_system, fs_base_path = infer_filesystem_type(remote["path"])

        with ProgressLogger(
            group="Upload",
            text="Uploading data artifacts",
            spinner="dots",
            on_success="Successfully uploaded model artifacts",
        ) as spinner:
            upload_assets(
                file_system=file_system,
                fs_base_path=fs_base_path,
                local_path=Path(
                    project_directory, "data", dataset["channel"]
                ).as_posix(),
                storage_location=Path("data", dataset["remote_path"]).as_posix(),
                zip_artifacts=dataset.get("compression", None) == "zip",
            )
            spinner.stop()

    except Exception as exception:
        logger.error(exception)
        raise


@click.command()
@click.option(
    "--channel",
    help="asset channel name. Directory name, within project data/ to store assets",
    required=True,
    type=str,
)
@click.option(
    "--name",
    help="asset filename name. File name, within project data/<channel> in which data artifact will be found",
    required=True,
    type=str,
)
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
def pull(channel, name, project_directory):
    """
    Command to create dataset manifest for mldock enabled container projects.
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception(
                (
                    "Path '{}' was not an mldock project. "
                    "Confirm this directory is correct, otherwise "
                    "create one.".format(project_directory)
                )
            )

        mldock_manager = MLDockConfigManager(
            filepath=Path(project_directory, MLDOCK_CONFIG_NAME)
        )

        # get mldock_module_dir name
        mldock_config = mldock_manager.get_config()

        input_data_channels = InputDataConfigManager(
            config=mldock_config.get("data", []),
            base_path=Path(project_directory, "data"),
        )

        dataset = input_data_channels.get(channel=channel, filename=name)

        config_manager = CliConfigureManager()

        remote = config_manager.remotes.get(name=dataset["remote"])

        file_system, fs_base_path = infer_filesystem_type(remote["path"])

        with ProgressLogger(
            group="Download",
            text="Downloading data artifacts",
            spinner="dots",
            on_success="Successfully downloaded model artifacts",
        ) as spinner:

            download_assets(
                file_system=file_system,
                fs_base_path=fs_base_path,
                storage_location=Path("data", dataset["remote_path"]).as_posix(),
                local_path=Path(
                    project_directory, "data", dataset["channel"]
                ).as_posix(),
            )
            spinner.stop()

    except Exception as exception:
        logger.error(exception)
        raise


def add_commands(cli_group: click.group):
    """
    add commands to cli group
    args:
        cli (click.group)
    """
    cli_group.add_command(create)
    cli_group.add_command(update)
    cli_group.add_command(remove)
    cli_group.add_command(push)
    cli_group.add_command(pull)


add_commands(datasets)



"""Predict Request API utilties"""
import io
import json
import csv
from pathlib import Path
import requests
from PIL import Image


def execute_request(url, headers, data):
    """compiles and executes request to API"""
    return requests.post(url=url, headers=headers, data=data)


def save_bytes_as_png(bytes_img, filepath):
    """save bytes image as png throgh pillow"""
    image = Image.open(io.BytesIO(bytes_img)).convert("RGB")

    image.save(filepath)


def save_json(json_data, filepath):
    """save json data to .json"""
    with open(filepath, "w") as file_:
        json.dump(json_data, file_)


def save_csv(text_data, filepath):
    """save text data as csv/txt"""
    with open(filepath, "w") as file_:
        writer = csv.writer(file_)

        for row in text_data:
            writer.writerow(row)


def send_json(data, host, **kwargs):
    """Send a json payload in request providing a filepath."""

    headers = kwargs.get("headers", None)
    if headers is None:
        headers = {}

    headers.update({"Content-Type": "application/json"})

    response = execute_request(url=host, headers=headers, data=json.dumps(data))
    if response.status_code != 200:
        raise requests.exceptions.RequestException(
            "error ({}): {}".format(response.status_code, response.raise_for_status())
        )

    return response


def send_csv(data, host, **kwargs):
    """Send a json payload in request providing a filepath."""

    headers = kwargs.get("headers", None)
    if headers is None:
        headers = {}

    headers.update({"Content-Type": "text/csv"})

    response = execute_request(url=host, headers=headers, data=data)

    if response.status_code != 200:
        raise requests.exceptions.RequestException(
            "error ({}): {}".format(response.status_code, response.raise_for_status())
        )
    return response


def send_image_jpeg(data, host, **kwargs):
    """Send a image jpeg payload as bytes in request providing a filepath."""

    headers = kwargs.get("headers", None)
    if headers is None:
        headers = {}

    headers.update({"Content-Type": "image/jpeg"})

    response = execute_request(url=host, headers=headers, data=data)

    if response.status_code != 200:
        raise requests.exceptions.RequestException(
            "error ({}): {}".format(response.status_code, response.raise_for_status())
        )
    return response


def handle_request(host, file_path, content_type, **kwargs):
    """handles and sends the request"""
    if content_type == "application/json":

        with open(file_path, "r") as file_:
            data = json.load(file_)

        response_obj = send_json(data, host, **kwargs)
        return response_obj

    elif content_type == "image/jpeg":

        with open(file_path, "rb") as file_:
            data = io.BytesIO(file_.read())

        response_obj = send_image_jpeg(data, host, **kwargs)

        return response_obj

    elif content_type == "text/csv":

        with open(file_path, "r") as file_:
            data = file_.read()

        response_obj = send_csv(data, host, **kwargs)

        return response_obj


def handle_response(response, content_type, file_path=None):
    """handle response. Extract, transform and emit/write to file"""

    if content_type == "application/json":

        if file_path is None:
            return response.json()

        else:
            save_json(response.json(), file_path)

    elif content_type == "image/jpeg":

        if file_path is None:
            return response._content

        else:
            save_bytes_as_png(response._content, file_path)

    elif content_type == "text/csv":

        if file_path is None:
            return response.text

        else:
            save_csv(response.text, file_path)


def handle_prediction(
    host: str,
    request: str,
    response_file: str = None,
    request_content_type: str = "application/json",
    response_content_type: str = "application/json",
    **kwargs
):
    """handles prediction workflow from handling the request to handling the response"""
    # handle request (load to obj, convert, send, get response)
    response_obj = handle_request(
        host=host,
        file_path=request,
        content_type=request_content_type,
        headers=kwargs.get("headers"),
    )

    # handle reponse (write file, print to terminal)
    return handle_response(
        response=response_obj,
        content_type=response_content_type,
        file_path=response_file,
    )



"""
    Docker Auth Utilities aand workflows
"""
import logging
import docker

from mldock.platform_helpers.aws.auth import get_aws_ecr
from mldock.platform_helpers.gcp.auth import get_gcp_gcr
from mldock.platform_helpers.dockerhub.auth import get_dockerhub_credentials

logger = logging.getLogger("mldock")


def login_and_authenticate(provider: str, region: str):
    """
    Login and authenticate with registry
    """
    client = docker.from_env()

    if provider == "ecr":
        # AWS ECR
        username, password, registry, repo = get_aws_ecr(region=region)
    elif provider == "gcr":
        # GCP
        username, password, registry, repo = get_gcp_gcr(region=region)
    elif provider == "dockerhub":
        # DockerHub
        username, password, registry, repo = get_dockerhub_credentials()
    else:
        raise TypeError(
            "{PROVIDER} Registry is not implemented.".format(PROVIDER=provider)
        )

    # login
    _ = client.login(
        username=username, password=password, email=None, registry=registry, reauth=True
    )

    # return logged in client
    return client, {
        "username": username,
        "password": password,
        "registry": registry,
        "repository": repo,
    }



"""
    TRAINER SCRIPT - contains custom code for training your model
"""
import os
import argparse
import pickle

import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC

from src.container.lifecycle import training_container


@training_container.wrap
def run_training(environment, logger):
    """Run training worflow"""
    parser = argparse.ArgumentParser()
    # get parameters
    parser.add_argument(
        "--iris",
        type=str,
        default=os.path.join(environment.input_data_dir, "iris"),
        help="Expects a .csv file, representing the lookup.",
    )
    parser.add_argument(
        "--test_size",
        type=float,
        default=environment.hyperparameters.get("test_size", 0.33),
        help="size of test sample to evaluate on.",
    )
    # Get args
    args, _ = parser.parse_known_args()

    data = pd.read_csv(os.path.join(args.iris, "train.csv"))
    X, y = data.drop(["target"], axis=1), data["target"]
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=args.test_size, random_state=123, shuffle=True
    )

    transformer = StandardScaler()
    transformer.fit(X_train)
    X_train_std = transformer.transform(X_train)
    X_test_std = transformer.transform(X_test)

    logger.info("Running Training")

    model = SVC(kernel="rbf", random_state=0, gamma=0.10, C=1.0)
    model.fit(X_train_std, y_train)

    logger.info("Running Evaluation")
    logger.info(f"train_accuracy={model.score(X_train_std, y_train)};")
    logger.info(f"test_accuracy={model.score(X_test_std, y_test)};")

    logger.info("Save Model artifacts")
    transformer_path = os.path.join(environment.model_dir, "iris/transformer.pkl")

    with open(transformer_path, "wb") as file_:
        pickle.dump(transformer, file_)

    model_path = os.path.join(environment.model_dir, "iris/model.pkl")

    with open(model_path, "wb") as file_:
        pickle.dump(model, file_)



"""
    Config managers for Configure commands

    Config managers used in command/configure
"""
import sys
import json
import logging
from pathlib import Path
import yaml
import click
from mldock.terminal import ChoiceWithNumbers, style_dropdown
from mldock.config_managers.core import BaseConfigManager
from mldock.platform_helpers import utils

logger = logging.getLogger("mldock")
MLDOCK_CLI_CONFIG = "./.mldock/config"


class CliConfigureManager(BaseConfigManager):
    """CLI Configure Manager for mldock"""

    config = {}

    def __init__(self, filepath: str = MLDOCK_CLI_CONFIG, create: bool = False):
        super().__init__()
        self.filepath = filepath

        self.config = self.load_config(self.filepath, create=create)

    def reset(self, config_type):
        """reset given configuration by config key"""
        click.secho("Dropping {} configuration".format(config_type), bg="blue")
        self.config.pop(config_type, None)

    def check_if_exists_else_create(self, file_name: str, create: bool):
        """check that mldock fileexists"""
        if not self.file_exists(file_name):
            if create:
                # deal with possiblity of nested directory
                utils._mkdir(Path(file_name).parents[0])
                # create file
                self.touch(file_name)
            else:
                logger.error(
                    (
                        "No MLDOCK CLI config found at '{CONFIG_PATH}/'. "
                        "To create run: 'mldock configure init'".format(
                            CONFIG_PATH=Path(file_name).parents[0]
                        )
                    )
                )
                sys.exit(1)

    def setup_local_config(self):
        """setup and prompt user for local configuration"""
        click.secho("Setup local CLI configuration", bg="blue")
        if self.config.get("local") is None:
            self.config["local"] = {}
        self.ask_for_container_auth_type()
        self.ask_for_environment(config_type="local")

    def setup_templates_config(self):
        """setup and prompt user for templates configuration"""
        click.secho("Setup templates CLI configuration", bg="blue")
        if self.config.get("templates") is None:
            self.config["templates"] = {}
        self.ask_for_template_server()
        self.ask_for_template_root()

    def setup_remotes_config(self):
        """setup and prompt user for templates configuration"""
        click.secho("Setup remotes CLI configuration", bg="blue")
        if self.config.get("remotes") is None:
            self.config["remotes"] = []
        self.ask_for_remote_path()

    def setup_workspace_config(self):
        """setup and prompt user for workspace configuration"""
        click.secho("Setup workspace CLI configuration", bg="blue")
        if self.config.get("workspace") is None:
            self.config["workspace"] = {}
        self.ask_for_environment(config_type="workspace")

    def ask_for_container_auth_type(self):
        """prompt user for platform name"""

        click.secho("Set local authentication type", bg="blue", nl=True)
        options = ["gcloud", "awscli"]

        auth_type = click.prompt(
            text=style_dropdown(
                group_name="authentication type",
                options=options,
                default=self.config.get("auth_type", None),
            ),
            type=ChoiceWithNumbers(options, case_sensitive=False),
            show_default=True,
            default=self.config.get("auth_type", None),
            show_choices=False,
        )

        self.config["local"].update({"auth_type": auth_type})

    def ask_for_environment(self, config_type):
        """
        prompt user for environment variables

        args:
            config_type (str): config key that points to configure config type
        """
        config = self.config[config_type].get("environment", {})
        environment_config_manager = EnvironmentConfigManager(config=config)
        environment_config_manager.ask_for_env_vars()
        self.config[config_type].update(
            {"environment": environment_config_manager.get_config()}
        )

    def ask_for_template_server(self):
        """prompt user to set template server tracking"""

        click.secho("Set template server type", bg="blue", nl=True)
        options = ["local", "github"]

        server_type = click.prompt(
            text=style_dropdown(
                group_name="template server type",
                options=options,
                default=self.config.get("server_type", None),
            ),
            type=ChoiceWithNumbers(options, case_sensitive=False),
            show_default=True,
            default=self.config.get("server_type", None),
            show_choices=False,
        )

        self.config["templates"].update({"server_type": server_type})

    def ask_for_template_root(self):
        """prompt user to set template root"""

        click.secho("Set template root", bg="blue", nl=True)

        templates_root_dir = click.prompt(
            text=click.style("Set path to template root dir: ", fg="bright_blue"),
            default=self.config["templates"].get("templates_root", None),
        )

        self.config["templates"].update({"templates_root": templates_root_dir})

    def ask_for_remote_path(self):
        """prompt user to set template root"""

        click.secho("Set remote path", bg="blue", nl=True)

        remote_path = click.prompt(
            text=click.style("Set path to remote: ", fg="bright_blue")
        )

        _, path = utils.get_bucket_and_path_from_scheme(remote_path)
        name = path.replace("/", " ").strip().replace(" ", ":")
        scheme = utils.get_scheme(remote_path)

        remotes_manager = RemotesConfigManager(config=self.config.get("remotes", []))

        remotes_manager.add_remote(name=name, type=scheme, path=remote_path)
        self.config["remotes"] = remotes_manager.get_config()

    @staticmethod
    def _format_nodes(config):

        output = []

        for key_ in sorted(config):
            value_ = config[key_]
            if isinstance(value_, dict):
                value_ = yaml.dump(value_, indent=4, sort_keys=True)
            output.append({"name": key_, "message": value_})

        return output

    def get_state(self):
        """pretty prints a json config to terminal"""
        config = self.config.copy()

        states = self._format_nodes(config)

        return states

    @property
    def local(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get("local", {})

    @property
    def workspace(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get("workspace", {})

    @property
    def templates(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get("templates", {})

    @property
    def remotes(self) -> dict:
        """get config object

        Returns:
            dict: config
        """

        return RemotesConfigManager(config=self.config.get("remotes", {}))


class PackageConfigManager(BaseConfigManager):
    """Package Requirement Config Manager for sagify"""

    def __init__(self, filepath: str, create: bool = False):
        super().__init__()
        self.filepath = filepath

        self.config = self.load_config(self.filepath, create=create)

    @staticmethod
    def touch(path):
        """creat an empty txt file

        Args:
            filename (str): path to file
        """
        Path(path).touch()

    def load_config(self, file_name: str, create: bool) -> list:
        """load config

        Args:
            filename (str): path to config file to load

        Returns:
            list: config
        """
        self.check_if_exists_else_create(file_name=file_name, create=create)
        path = Path(file_name)
        with path.open() as file_:
            config = file_.readlines()

        for index, c_package in enumerate(config):
            if c_package.endswith("\n"):
                config[index] = config[index].split("\n")[0]
        return config

    def get_config(self) -> list:
        """get config object

        Returns:
            list: config
        """
        return self.config

    def seed_packages(self, packages: list):
        """seeds config with required package modules

        Args:
            packages (list): packages to add to config
        """
        for new_package in packages:
            for config_package in self.config:
                if new_package not in config_package:
                    self.config.append(new_package)

    def pretty_print(self):
        """pretty prints a list config to terminal"""
        pretty_config = json.dumps(self.config, indent=4, separators=(",", ": "))
        logger.debug("{}\n".format(pretty_config))

    def write_file(self):
        """
        write to file
        """
        config = set(self.config)
        config_txt = "\n".join(config) + "\n"
        Path(self.filepath).write_text(config_txt)


class StageConfigManager(BaseConfigManager):
    """Development Stage Config Manager for mldock"""

    def __init__(self, config: dict):
        self.config = config

    def ask_for_stages(self):
        """prompt user for stages"""
        click.secho("Stages", bg="blue", nl=True)
        while True:
            echo_msg = (
                click.style("Add a development stage. ", fg="bright_blue")
                + "(Follow the prompts)"
            )

            click.echo(echo_msg, nl=True)

            stage_name = click.prompt(
                text="Stage name: ", default="end", show_default=False, type=str
            )
            if stage_name == "end":
                break

            docker_tag = click.prompt(
                text="Set docker image tag: ",
                default=self.config.get(stage_name, {}).get("tag", None),
            )
            if stage_name not in self.config:
                self.config[stage_name] = {}

            self.config[stage_name].update({"tag": docker_tag})


class HyperparameterConfigManager(BaseConfigManager):
    """Hyperparameter Config Manager for mldock"""

    def __init__(self, config: dict):
        self.config = config

    @staticmethod
    def is_float(value):
        """Returns True is string is a number."""
        try:
            float(value)
            return True
        except ValueError:
            return False

    def ask_for_hyperparameters(self):
        """prompt user for hyperparameters"""
        click.secho("Hyperparameters", bg="blue", nl=True)
        while True:

            echo_msg = (
                click.style("Add a hyperparameter. ", fg="bright_blue")
                + "(Follow prompts)"
            )

            click.echo(echo_msg, nl=True)
            hparam_name = click.prompt(
                text="Hyperparameter (name): ",
                default="end",
                show_default=False,
                type=str,
            )

            if hparam_name == "end":
                break

            hparam_value = click.prompt(
                text="Set value: ", default=self.config.get(hparam_name, None)
            )

            self.config.update({hparam_name: hparam_value})


class EnvironmentConfigManager(BaseConfigManager):
    """Environment Config Manager for mldock"""

    def __init__(self, config: dict):
        self.config = config

    @staticmethod
    def is_float(value):
        """Returns True is string is a number."""
        try:
            float(value)
            return True
        except ValueError:
            return False

    @staticmethod
    def parse_bool(value):
        """Returns True is string is a bool."""
        result = value
        if value.lower() == "true":
            result = True
        elif value.lower() == "false":
            result = False
        return result

    def ask_for_env_vars(self):
        """prompt user for environment variables"""
        click.secho("Environment Variables", bg="blue", nl=True)
        while True:
            echo_msg = (
                click.style("Add a environment variable. ", fg="bright_blue")
                + "(Follow prompts)"
            )

            click.echo(echo_msg, nl=True)

            env_var_name = click.prompt(
                text="Environment Variable (name): ",
                default="end",
                show_default=False,
                type=str,
            )
            if env_var_name == "end":
                break

            env_var_value = click.prompt(
                text="Set value: ", default=self.config.get(env_var_name, None)
            )

            self.config.update({env_var_name: env_var_value})


class InputDataConfigManager(BaseConfigManager):
    """InputData Config Manager for mldock"""

    file_ignores = []

    def __init__(self, config: dict, base_path: str):
        self.config = config
        self.base_path = base_path

    def write_gitignore(self):
        """
        write to file
        """
        config = {
            Path(dataset["channel"], dataset["filename"]).as_posix()
            for dataset in self.config
        }

        config_txt = "\n".join(config) + "\n"

        ignore_filepath = Path(self.base_path, ".gitignore")
        Path(ignore_filepath).write_text(config_txt)

    def check_if_asset_exists(self, channel, filename):

        found_match = False
        for i in range(len(self.config)):
            tmp_config = self.config[i]

            if (tmp_config["channel"] == channel) and (
                tmp_config["filename"] == filename
            ):
                # pop the current config
                found_match = True

                return found_match, i
        return False, None

    def find_all_for_given_channel(self, channel):

        datasets = []
        for i in range(len(self.config)):
            tmp_config = self.config[i]

            if (tmp_config["channel"] == channel):

                datasets.append(tmp_config)

        return datasets

    def add_asset(self, update: bool = False, **data_config):
        """add an asset to data config"""

        found_match, i = self.check_if_asset_exists(
            channel=data_config.get("channel"), filename=data_config.get("filename")
        )

        if (update == False) and found_match:
            logger.error(
                "dataset artifact already exists. To update, run 'datasets update' instead."
            )
            exit(0)
        elif update and found_match:
            current_config = self.config.pop(i)
            current_config.update(data_config)
        elif update and (found_match == False):
            logger.error(
                "dataset artifact does not exists. To create, run 'datasets create' instead."
            )
            exit(0)
        else:
            current_config = data_config

        self.config.append(data_config)

    def remove(self, **data_config):
        found_match, i = self.check_if_asset_exists(
            channel=data_config.get("channel"), filename=data_config.get("filename")
        )

        if found_match:
            self.config.pop(i)
        else:
            logger.error("dataset artifact does not exists, so nothing to remove.")

    def get(self, **data_config):
        found_match, i = self.check_if_asset_exists(
            channel=data_config["channel"], filename=data_config["filename"]
        )

        if found_match:
            return self.config[i]
        else:
            logger.error(
                "dataset artifact does not exists. To create, run 'datasets create' instead."
            )
            exit(0)

    def get_all(self, **data_config):
        datasets = self.find_all_for_given_channel(
            channel=data_config["channel"]
        )

        if len(datasets) > 0:
            return datasets
        else:
            logger.error(
                "dataset artifact does not exists. To create, run 'datasets create' instead."
            )
            exit(0)

    def ask_for_input_data_channels(self):
        """prompt user for hyperparameters"""
        click.secho("Input Data Channels", bg="blue", nl=True)
        while True:

            echo_msg = (
                click.style("Add a data channel. ", fg="bright_blue")
                + "(Expects channel/filename). Hit enter to continue."
            )

            channel_filename_pair = click.prompt(
                text=echo_msg, default="end", show_default=False, type=str
            )

            # pylint: disable=no-else-break
            if channel_filename_pair == "end":

                logger.debug("\nUpdated data channels")

                self.pretty_print()

                break

            elif (
                "/" in channel_filename_pair
                and channel_filename_pair != "channel/filename"
            ):

                channel, filename = channel_filename_pair.split("/", 1)

                logger.debug(
                    (
                        "Adding data/{CHANNEL}/{FILE_NAME}".format(
                            CHANNEL=channel, FILE_NAME=filename
                        )
                    )
                )

                self.config.append({"channel": channel, "filename": filename})

            else:
                logger.warning("Expected format as channel/filename. Skipping")


class ModelConfigManager(BaseConfigManager):
    """Model Config Manager for mldock"""

    file_ignores = []

    def __init__(self, config: dict, base_path: str):
        self.config = config
        self.base_path = base_path

    def write_gitignore(self):
        """
        write to file
        """
        config = {
            Path(dataset["channel"], dataset["filename"]).as_posix()
            for dataset in self.config
        }

        config_txt = "\n".join(config) + "\n"

        ignore_filepath = Path(self.base_path, ".gitignore")
        Path(ignore_filepath).write_text(config_txt)

    def check_if_asset_exists(self, channel, filename):

        found_match = False
        for i in range(len(self.config)):
            tmp_config = self.config[i]

            if (tmp_config["channel"] == channel) and (
                tmp_config["filename"] == filename
            ):
                # pop the current config
                found_match = True

                return found_match, i
        return False, None

    def find_all_for_given_channel(self, channel):

        datasets = []
        for i in range(len(self.config)):
            tmp_config = self.config[i]

            if (tmp_config["channel"] == channel):

                datasets.append(tmp_config)

        return datasets

    def add_asset(self, update: bool = False, **data_config):
        """add an asset to data config"""

        found_match, i = self.check_if_asset_exists(
            channel=data_config.get("channel"), filename=data_config.get("filename")
        )

        if (update == False) and found_match:
            logger.error(
                "model artifact already exists. To update, run 'datasets update' instead."
            )
            exit(0)
        elif update and found_match:
            current_config = self.config.pop(i)
            current_config.update(data_config)
        elif update and (found_match == False):
            logger.error(
                "model artifact does not exists. To create, run 'models create' instead."
            )
            exit(0)
        else:
            current_config = data_config

        self.config.append(data_config)

    def remove(self, **data_config):
        found_match, i = self.check_if_asset_exists(
            channel=data_config.get("channel"), filename=data_config.get("filename")
        )

        if found_match:
            self.config.pop(i)
        else:
            logger.error("model artifact does not exists, so nothing to remove.")

    def get(self, **model_config):
        found_match, i = self.check_if_asset_exists(
            channel=model_config["channel"], filename=model_config["filename"]
        )

        if found_match:
            return self.config[i]
        else:
            logger.error(
                "model artifact does not exists. To create, run 'models create' instead."
            )
            exit(0)

    def get_all(self, **model_config):
        models = self.find_all_for_given_channel(
            channel=model_config["channel"]
        )

        if len(models) > 0:
            return models
        else:
            logger.error(
                "dataset artifact does not exists. To create, run 'datasets create' instead."
            )
            exit(0)

    def ask_for_model_channels(self):
        """prompt user for hyperparameters"""
        click.secho("Model Channels", bg="blue", nl=True)
        while True:

            echo_msg = (
                click.style("Add a model. ", fg="bright_blue")
                + "(Expects channel/filename). Hit enter to continue."
            )

            channel_filename_pair = click.prompt(
                text=echo_msg, default="end", show_default=False, type=str
            )

            # pylint: disable=no-else-break
            if channel_filename_pair == "end":

                logger.debug("\nUpdated model channels")

                self.pretty_print()

                break

            elif (
                "/" in channel_filename_pair
                and channel_filename_pair != "channel/filename"
            ):

                channel, filename = channel_filename_pair.split("/", 1)

                logger.debug(
                    (
                        "Adding model/{CHANNEL}/{FILE_NAME}".format(
                            CHANNEL=channel, FILE_NAME=filename
                        )
                    )
                )

                self.config.append({"channel": channel, "filename": filename})

            else:
                logger.warning("Expected format as channel/filename. Skipping")


class RemotesConfigManager(BaseConfigManager):
    """Remotes Config Manager for mldock"""

    file_ignores = []

    def __init__(self, config: dict):
        self.config = config

    def check_if_asset_exists(self, name):

        found_match = False
        for i in range(len(self.config)):
            tmp_config = self.config[i]

            if tmp_config["name"] == name:
                # pop the current config
                found_match = True

                return found_match, i
        return False, None

    def add_remote(self, update: bool = False, **data_config):
        """add an asset to data config"""

        if data_config.get("path", None) is None:
            self.ask_for_remote_path()

        found_match, i = self.check_if_asset_exists(name=data_config.get("name"))

        if (update == False) and found_match:
            logger.error(
                "remote already exists. Check your .mldock/config file to view and edit remotes."
            )
            exit(0)
        elif update and found_match:
            current_config = self.config.pop(i)
            current_config.update(data_config)
        elif update and (found_match == False):
            logger.error(
                "remote does not exists. Check your .mldock/config file to view and edit remotes."
            )
            exit(0)
        else:
            current_config = data_config

        self.config.append(data_config)

    def remove(self, **data_config):
        found_match, i = self.check_if_asset_exists(channel=data_config["name"])

        if found_match:
            self.config.pop(i)
        else:
            logger.error(
                "remote does not exists. Check your .mldock/config file to view and edit remotes."
            )

    def get(self, **data_config):
        found_match, i = self.check_if_asset_exists(name=data_config["name"])

        if found_match:
            return self.config[i]
        else:
            logger.error(
                "remote does not exists. Check your .mldock/config file to view and edit remotes."
            )
            exit(0)



"""
    GCP Auth utilities
"""
import os
from google.auth import default
from google.auth.transport.requests import Request

from mldock.platform_helpers.utils import strip_scheme


def get_gcp_gcr(region: str):
    """
    Get Authentication credentials for google.auth

    args:
        region (str): Region to Authenticate GCR registry in

    return:
        username (str): username to use in docker client auth
        password (str): password token to use in docker client auth
        registry (str): registry url to use
    """
    credentials, project = default(
        scopes=["https://www.googleapis.com/auth/cloud-platform"]
    )
    # creds.valid is False, and creds.token is None
    # Need to refresh credentials to populate those
    auth_req = Request()
    credentials.refresh(auth_req)
    # set docker creds
    username = "oauth2accesstoken"
    password = credentials.token
    if region == "eu":
        registry = "https://eu.gcr.io"
    elif region == "us":
        registry = "https://us.gcr.io"
    elif region == "asia":
        registry = "https://asia.gcr.io"
    else:
        registry = "https://gcr.io"

    registry = os.path.join(registry, project)
    # return docker credentials
    cloud_repository = strip_scheme(registry)
    # return docker credentials
    return username, password, registry, cloud_repository



"""Logger formatter"""
# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import sys
import logging
import click

PY2 = sys.version_info[0] == 2


class ColorFormatter(logging.Formatter):
    """ColorFormatter for logging.Logger"""

    colors = {
        "error": dict(fg="red"),
        "exception": dict(fg="red"),
        "critical": dict(fg="red"),
        "debug": dict(fg="blue"),
        "warning": dict(fg="yellow"),
    }

    def format(self, record):
        """format record method"""
        if not record.exc_info:
            level = record.levelname.lower()
            msg = record.msg
            if level in self.colors:
                prefix = click.style("{}: ".format(level), **self.colors[level])
                if not PY2 and isinstance(msg, bytes):
                    msg = msg.decode(sys.getfilesystemencoding(), "replace")
                elif not isinstance(msg, (str, bytes)):
                    msg = str(msg)
                msg = "\n".join(prefix + x for x in msg.splitlines())
            return msg
        return logging.Formatter.format(self, record)


class ClickHandler(logging.Handler):
    """Click log format handler based on logging.Handler"""

    def emit(self, record):
        """log record emitter"""
        # pylint: disable=broad-except
        try:
            msg = self.format(record)
            level = record.levelname.lower()
            err = level in ("warning", "error", "exception", "critical")
            click.echo(msg, err=err)
        except Exception:
            self.handleError(record)


def _normalize_logger(logger, log_level):
    """normalize logger based on log_level."""
    if not isinstance(logger, logging.Logger):
        logger = logging.getLogger(logger)

    logger.setLevel(log_level)
    return logger


def configure_logger(logger, verbose=False):
    """configure a new logger in either DEBUG or INFO based on verbose flag."""
    log_level = logging.DEBUG if verbose else logging.INFO

    logger = _normalize_logger(logger, log_level)
    _default_handler = ClickHandler()
    _default_handler.formatter = ColorFormatter()
    logger.handlers = [_default_handler]
    logger.propagate = False
    return logger



"""CONFIGURE CLI COMMANDS"""
from pathlib import Path
import logging
import click

from mldock.config_managers.cli import CliConfigureManager

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")


def reset_terminal():
    """clears the terminal view frame"""
    click.clear()


@click.group()
def configure():
    """
    Commands to configure mldock cli.
    """


@click.command()
def init():
    """Configure  development tools"""
    config_manager = CliConfigureManager(create=True)
    config_manager.write_file()


@click.command()
@click.option(
    "-c",
    "--config-name",
    help="Configuration name you wish to reset",
    required=True,
    type=click.Choice(
        ["local", "workspace", "templates", "all", "remotes"], case_sensitive=False
    ),
)
@click.pass_obj
def reset(obj, config_name):
    """reset configurations"""
    config_manager = CliConfigureManager()
    if config_name == "all":
        for config in ["local", "workspace", "templates", "remotes"]:
            config_manager.reset(config)
    else:
        config_manager.reset(config_name)
    config_manager.write_file()

    reset_terminal()
    logger.info(obj["logo"])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg="blue"), nl=True)
        click.echo(click.style(state["message"], fg="white"), nl=True)


@click.command()
@click.pass_obj
def workspace(obj):
    """Configure for local development tools"""

    config_manager = CliConfigureManager(create=True)
    config_manager.setup_workspace_config()
    config_manager.write_file()

    reset_terminal()
    logger.info(obj["logo"])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg="blue"), nl=True)
        click.echo(click.style(state["message"], fg="white"), nl=True)


@click.command()
@click.pass_obj
def local(obj):
    """Configure for local development tools"""

    config_manager = CliConfigureManager(create=True)
    config_manager.setup_local_config()
    config_manager.write_file()

    reset_terminal()
    logger.info(obj["logo"])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg="blue"), nl=True)
        click.echo(click.style(state["message"], fg="white"), nl=True)


@click.command()
@click.pass_obj
def templates(obj):
    """Configure for local development tools"""

    config_manager = CliConfigureManager(create=True)
    config_manager.setup_templates_config()
    config_manager.write_file()

    reset_terminal()
    logger.info(obj["logo"])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg="blue"), nl=True)
        click.echo(click.style(state["message"], fg="white"), nl=True)


@click.command()
@click.pass_obj
def remotes(obj):
    """Configure for remotes development tools"""

    config_manager = CliConfigureManager(create=True)
    config_manager.setup_remotes_config()
    config_manager.write_file()

    reset_terminal()
    logger.info(obj["logo"])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg="blue"), nl=True)
        click.echo(click.style(state["message"], fg="white"), nl=True)


@click.command()
def show():
    """Configure for local development tools"""
    try:
        config_manager = CliConfigureManager()

        states = config_manager.get_state()

        for state in states:
            click.echo(click.style(state["name"], bg="blue"), nl=True)
            click.echo(click.style(state["message"], fg="white"), nl=True)
    except FileNotFoundError as exception:
        if Path(exception.filename).name == "mldock":
            logger.error(
                "File not found. Please run 'mldock configure local' "
                "to generate cli configuration."
            )
        else:
            raise


def add_commands(cli_group: click.group):
    """
    add commands to cli group
    args:
        cli (click.group)
    """
    cli_group.add_command(init)
    cli_group.add_command(workspace)
    cli_group.add_command(local)
    cli_group.add_command(templates)
    cli_group.add_command(remotes)
    cli_group.add_command(show)
    cli_group.add_command(reset)


add_commands(configure)



"""BASE ENVIRONMENT INTERFACE"""
import os
from pathlib import Path
import logging
import environs

from mldock.platform_helpers import utils


logger = logging.getLogger("mldock")


class BaseEnvironment:
    """Base Environment to be extended for different environments"""

    # pylint: disable=too-many-instance-attributes
    input_channel_regex = None
    model_channel_regex = None
    output_channel_regex = None
    output_channel_regex = None
    hyperparameters_file = None
    environment_variables = environs.Env()

    def __init__(self, base_dir: str = None, **kwargs):

        self.environment_variables.read_env()
        self.base_dir = base_dir
        if self.base_dir is None:
            # look for it in environment
            # default to /opt/ml
            self.base_dir = self.environment_variables.str("MLDOCK_BASE_DIR", "/opt/ml")

        self.relative_input_dir = self.environment_variables.str(
            "MLDOCK_INPUT_DIR", "input"
        )
        if self.relative_input_dir is None:
            self.relative_input_dir = "input"

        self.hyperparameters_file = kwargs.get(
            "hyperparameters_file", "hyperparameters.json"
        )

        self.input_channel_regex = kwargs.get(
            "input_channel_regex", r"MLDOCK_INPUT_CHANNEL_.*"
        )
        self.model_input_channel_regex = kwargs.get(
            "model_input_channel_regex", r"MLDOCK_MODEL_INPUT_CHANNEL.*"
        )

        self.model_output_channel_regex = kwargs.get(
            "model_output_channel_regex", r"MLDOCK_MODEL_OUTPUT_CHANNEL.*"
        )

        self.output_channel_regex = kwargs.get(
            "output_channel_regex", r"MLDOCK_OUTPUT_CHANNEL_.*"
        )
        self.hyperparamters_env_variable = kwargs.get(
            "hyperparamters_regex", r"MLDOCK_HYPERPARAMETERS"
        )

        self._create_training_directories()
        self.setup_hyperparameters()

    def _create_training_directories(self):
        """Create the directory structure, if not exists"""
        logger.debug(
            (
                "Creating a new training folder under {BASE_DIR} .".format(
                    BASE_DIR=self.base_dir
                )
            )
        )

        self.input_data_dir.mkdir(parents=True, exist_ok=True)
        self.model_dir.mkdir(parents=True, exist_ok=True)
        self.input_config_dir.mkdir(parents=True, exist_ok=True)
        self.output_data_dir.mkdir(parents=True, exist_ok=True)
        if not self.hyperparameters_filepath.exists():
            utils._write_json(obj={}, file_path=self.hyperparameters_filepath)

    @property
    def input_dir(self):
        """path to input directory in working directory"""
        return Path(self.base_dir, self.relative_input_dir)

    @property
    def input_data_dir(self):
        """path to input data directory in working directory"""
        return Path(self.input_dir, "data")

    @property
    def input_config_dir(self):
        """path to input config directory in working directory"""
        return Path(self.input_dir, "config")

    @property
    def hyperparameters_filepath(self):
        """path to hyperparameter config in input config directory in working directory"""
        return Path(self.input_config_dir, self.hyperparameters_file)

    @property
    def model_dir(self):
        """path to model directory in working directory"""
        return Path(self.base_dir, "model")

    @property
    def output_data_dir(self):
        """path to output directory in working directory"""
        return Path(self.base_dir, "output")

    @property
    def hyperparameters(self):
        """Returns: Iterable of the input channels"""
        return utils._read_json(self.hyperparameters_filepath)

    def setup_hyperparameters(self):
        """Retrieves the env vars matching hyperparameters regex and updates config"""
        hyperparameters = utils._read_json(self.hyperparameters_filepath.as_posix())

        updated_hparams = self.environment_variables.json(
            self.hyperparamters_env_variable, "{}"
        )

        hyperparameters.update(updated_hparams)

        utils._write_json(
            obj=hyperparameters, file_path=self.hyperparameters_filepath.as_posix()
        )

    def get_input_channel_iter(self):
        """Returns: Iterable of the input channels"""
        envvars = dict(os.environ)
        return utils.get_env_vars(environment=envvars, regex=self.input_channel_regex)

    def get_output_channel_iter(self):
        """Returns: Iterable of the output channel"""
        envvars = dict(os.environ)
        return utils.get_env_vars(environment=envvars, regex=self.output_channel_regex)

    def get_model_input_channel_iter(self):
        """Returns: Iterable of the model channel"""
        envvars = dict(os.environ)
        return utils.get_env_vars(
            environment=envvars, regex=self.model_input_channel_regex
        )

    def get_model_output_channel_iter(self):
        """Returns: Iterable of the model channel"""
        envvars = dict(os.environ)
        return utils.get_env_vars(
            environment=envvars, regex=self.model_output_channel_regex
        )

    @staticmethod
    def setup_inputs():
        """Iterates and downloads assets remoate -> input channels
        return:
            None
        """
        # pylint: disable=unnecessary-pass
        pass

    @staticmethod
    def cleanup_outputs():
        """Iterates and uploads output channel -> remote
        return:
            None
        """
        # pylint: disable=unnecessary-pass
        pass

    @staticmethod
    def setup_model_artifacts():
        """Iterates and downloads assets remoate -> model channel
        return:
            None
        """
        # pylint: disable=unnecessary-pass
        pass

    @staticmethod
    def cleanup_model_artifacts():
        """Iterates and uploads from model channel -> remote
        return:
            None
        """
        # pylint: disable=unnecessary-pass
        pass



from pyarrow import fs
import gcsfs
import s3fs

from mldock.platform_helpers import utils


def infer_filesystem_type(path: str):
    """
    infers appropriate filesystem type and returns pyarrow support file-system

    args:
        path (str): full path/uri
    returns:
        file_system: a supported file-system type
        path: path without scheme
    """

    if utils._check_if_cloud_scheme(path, scheme=""):
        file_system = fs.LocalFileSystem()
    elif utils._check_if_cloud_scheme(path, scheme="s3"):
        path_without_scheme = utils.strip_scheme(path)
        file_system, path = s3fs.S3FileSystem(), path_without_scheme
    elif utils._check_if_cloud_scheme(path, scheme="gs"):
        path_without_scheme = utils.strip_scheme(path)
        file_system, path = gcsfs.GCSFileSystem(), path_without_scheme
    else:
        raise TypeError(
            "path scheme = '{SCHEME}' for '{PATH}' "
            "is not currently supported. "
            "Available options = 's3' or 'gs' "
            "or local filesystem path".format(SCHEME=utils.get_scheme(path), PATH=path)
        )
    return file_system, path



"""MLDock Friendly Environments for GCP"""
from pathlib import Path
import logging
import gcsfs

from mldock.platform_helpers.mldock.configuration.environment import base

# from mldock.platform_helpers.gcp.storage import \
#     download_input_assets, package_and_upload_model_dir, package_and_upload_output_data_dir
from mldock.platform_helpers.mldock.storage.pyarrow import (
    download_assets,
    upload_assets,
)
from mldock.platform_helpers import utils

logger = logging.getLogger("mldock")


class GCPEnvironment(base.BaseEnvironment):
    """
    Extends the Environment class to give us
    more specific environment configuration tasks based on
    the GCP resource eco-system.
    """

    def setup_inputs(self):
        """Iterates and downloads assets remoate -> input channels"""
        logger.debug(
            (
                "Setup assets in {INPUT_DATA_DIR}".format(
                    INPUT_DATA_DIR=self.input_data_dir
                )
            )
        )
        # only fetch channels of environment prefix MLDOCK_INPUT_CHANNEL_
        channels = self.get_input_channel_iter()

        if len(channels) == 0:
            logger.debug("No input channels were found in ENV VARS.")
        for channel in channels:
            channel_path = channel["key"].replace("MLDOCK_INPUT_CHANNEL_", "").lower()
            local_channel_path = Path(self.input_data_dir, channel_path)
            try:
                path_without_scheme = utils.strip_scheme(channel["value"])
                file_system = gcsfs.GCSFileSystem()
                download_assets(
                    file_system,
                    fs_base_path=path_without_scheme,
                    local_path=local_channel_path,
                    storage_location=".",
                )

            except FileExistsError:
                logger.debug(
                    (
                        "{CHANNEL_KEY} Channel skipped. Already exists".format(
                            CHANNEL_KEY=channel["key"]
                        )
                    )
                )

    def cleanup_outputs(self):
        """Iterates and uploads output channel -> remote"""
        logger.debug("Cleanup assets in {}".format(self.output_data_dir))
        # only fetch channels of environment prefix MLDOCK_OUTPUT_CHANNEL_
        channels = self.get_output_channel_iter()

        if len(channels) == 0:
            logger.debug("No output channels were found in ENV VARS.")

        for channel in channels:
            # only fetch channels with output
            channel_path = channel["key"].replace("MLDOCK_OUTPUT_CHANNEL_", "").lower()
            local_channel_path = Path(self.output_data_dir, channel_path)
            try:
                path_without_scheme = utils.strip_scheme(channel["value"])
                file_system = gcsfs.GCSFileSystem()
                upload_assets(
                    file_system,
                    fs_base_path=path_without_scheme,
                    local_path=local_channel_path,
                    storage_location=".",
                    zip_artifacts=True,
                )

            except AssertionError:
                logger.debug(
                    "Skipping channel. {LOCAL_CHANNEL_PATH} is not a "
                    "directory or could not be found".format(
                        LOCAL_CHANNEL_PATH=local_channel_path
                    )
                )

    def setup_model_artifacts(self):
        """Iterates and downloads assets remoate -> model channel"""
        logger.debug("Setup model assets in {}".format(self.model_dir))
        # only fetch channels of environment prefix MLDOCK_MODEL_INPUT_CHANNEL_
        channels = self.get_model_input_channel_iter()

        if len(channels) == 0:
            logger.debug("No input channels were found in ENV VARS.")

        for channel in channels:
            channel_path = (
                channel["key"].replace("MLDOCK_MODEL_INPUT_CHANNEL_", "").lower()
            )
            local_channel_path = Path(self.model_dir, channel_path)
            try:
                path_without_scheme = utils.strip_scheme(channel["value"])
                file_system = gcsfs.GCSFileSystem()
                download_assets(
                    file_system,
                    fs_base_path=path_without_scheme,
                    local_path=local_channel_path,
                    storage_location=".",
                )

            except FileExistsError:
                logger.debug(
                    (
                        "{CHANNEL_KEY} Channel skipped. Already exists".format(
                            CHANNEL_KEY=channel["key"]
                        )
                    )
                )

    def cleanup_model_artifacts(self):
        """Iterates and uploads from model channel -> remote"""
        logger.debug("Cleanup model assets in {}".format(self.model_dir))

        # only fetch channels of environment prefix MLDOCK_MODEL_OUTPUT_CHANNEL_
        channels = self.get_model_output_channel_iter()

        if len(channels) == 0:
            logger.debug("No model channels were found in ENV VARS.")

        for channel in channels:
            channel_path = (
                channel["key"].replace("MLDOCK_MODEL_OUTPUT_CHANNEL_", "").lower()
            )
            local_channel_path = Path(self.model_dir, channel_path)

            try:

                path_without_scheme = utils.strip_scheme(channel["value"])
                file_system = gcsfs.GCSFileSystem()
                upload_assets(
                    file_system,
                    fs_base_path=path_without_scheme,
                    local_path=local_channel_path,
                    storage_location=".",
                    zip_artifacts=True,
                )

            except AssertionError:
                logger.debug(
                    "Skipping channel. {LOCAL_CHANNEL_PATH} is not a "
                    "directory or could not be found".format(
                        LOCAL_CHANNEL_PATH=local_channel_path
                    )
                )



"""
    Decorators for model service workflows

    Provides a set of decorators for Training and prediction workflows, which include
    error handling and start up and cleanup steps in the workflow management.
"""
from functools import wraps
import os
import traceback


def make_decorator(foreign_decorator):
    """
    Returns a copy of foreignDecorator, which is identical in every
    way(*), except also appends a .decorator property to the callable it
    spits out.
    """

    def new_decorator(func):
        """wrapper to create the new decorated func"""
        decorator_result = foreign_decorator(func)
        decorator_result.decorator = new_decorator

        return decorator_result

    new_decorator.__name__ = foreign_decorator.__name__
    new_decorator.__doc__ = foreign_decorator.__doc__

    return new_decorator


def trainer(container, environment, logger):
    """
    Args:
        TrainingContainer, environment, logger
    """

    def general_decorator(function):
        @wraps(function)
        def func_wrapper(*args, **kwargs):
            """
            Wrapper function to decorate function

            This could be train.py:
            """
            training = container(
                container_environment=environment, container_logger=logger
            )
            try:
                training.startup()

                function(*args, **kwargs)
            except Exception as exception:
                # Write out an error file. This will be returned as the failureReason in the
                # DescribeTrainingJob result.
                trc = traceback.format_exc()
                log_file_path = os.path.join(environment.output_data_dir, "failure")
                with open(log_file_path, "w") as file_:
                    file_.write(
                        "Exception during training: " + str(exception) + "\n" + trc
                    )
                # Printing this causes the exception to be in the training job logs, as well.
                logger.error(
                    "Exception during training: " + str(exception) + "\n" + trc
                )
                raise
            finally:
                training.cleanup()

        return func_wrapper

    # add doc and decorator metadata
    general_decorator = make_decorator(general_decorator)
    return general_decorator


def predictor(container, environment, logger):
    """
    Args:
        TrainingContainer, environment, logger

    note:
        - this is called each time server calls your method, now decorated with this.
    """

    def general_decorator(function):
        @wraps(function)
        def func_wrapper(*args, **kwargs):
            """
            Wrapper function to decorate function

            This could be train.py:
            """
            serving = container(
                container_environment=environment, container_logger=logger
            )
            try:
                serving.startup_worker()

                return function(*args, **kwargs)
            except Exception as exception:
                # Write out an error file. This will be returned as the failureReason in the
                # DescribeTrainingJob result.
                trc = traceback.format_exc()
                log_file_path = os.path.join(environment.output_data_dir, "failure")
                with open(log_file_path, "w") as file_:
                    file_.write(
                        "Exception during training: " + str(exception) + "\n" + trc
                    )
                # Printing this causes the exception to be in the training job logs, as well.
                logger.error(
                    "Exception during training: " + str(exception) + "\n" + trc
                )
                raise
            finally:
                serving.cleanup_worker()

        return func_wrapper

    # add doc and decorator metadata
    general_decorator = make_decorator(general_decorator)
    return general_decorator



"""LOGS API UTILITIES"""
import logging
from pathlib import Path
from clickclick.console import print_table
from pygrok import Grok
from pyarrow import fs

from mldock.api.assets import infer_filesystem_type

logger = logging.getLogger("mldock")

STYLES = {
    "FINE": {"fg": "green"},
    "ERROR": {"fg": "red"},
    "WARNING": {"fg": "yellow", "bold": True},
}


TITLES = {
    "state": "Status",
    "creation_time": "Creation Date",
    "id": "Identifier",
    "desc": "Description",
    "name": "Name",
}

MAX_COLUMN_WIDTHS = {
    "desc": 50,
    "name": 20,
}


def read_file_stream(file_path, file_system):
    """
    Read file from file_system as a stream.

    args:
        file_path (str): path to file
        file_system (pyarrow.fs.FileSystem): pyarrow supported filesystem object
    return
        log_data (str): log data
    """
    if isinstance(file_system, fs.LocalFileSystem):
        with file_system.open_input_stream(file_path) as file_:
            log_data = file_.read().decode()
    else:
        with file_system.open(file_path, "rb") as file_:
            log_data = file_.read().decode()
    return log_data


def parse_grok(file_path, pattern, file_system):
    """
    Parse logs using a custom crok pattern and return parsed objects.

    args:
        file_path (str): file path to object in file_system
        pattern (str): supported grok pattern
        file_system (pyarrow.fs.FileSystem): pyarrow supported file system object
    return:
        metadata (list): list of file objects as per grok pattern
    """
    grok = Grok(pattern)

    log_data = read_file_stream(file_path, file_system)

    metadata = grok.match(log_data.replace("\n", "\\n"))

    return metadata


def parse_grok_multiline(file_path, pattern, file_system):
    """
    Parse logs using a custom crok pattern and return parsed objects.

    args:
        file_path (str): file path to object in file_system
        pattern (str): supported grok pattern
        file_system (pyarrow.fs.FileSystem): pyarrow supported file system object
    return:
        metadata (list): list of file objects as per grok pattern
    """
    grok = Grok(pattern)

    metadata = []

    log_data = read_file_stream(file_path, file_system)

    for log in log_data.split("\n"):

        result = grok.match(log)
        if result is not None:
            metadata.append(result)

    return metadata


def get_all_file_objects(base_path, file_name, file_system):
    """
    Get all file objects, filter and return all files matching file_name

    args:
        base_path (str): base path to start walk from to find file objects
        file_name (str): file names to match and return
        file_system (pyarrow.fs.FileSystem): pyarrow supported file system object
    return:
        file_paths (list): list of file paths
    """

    if isinstance(file_system, fs.LocalFileSystem):
        file_selector = fs.FileSelector(base_path, recursive=True)
        logs = [
            log.path
            for log in file_system.get_file_info(file_selector)
            if log.base_name == file_name
        ]
    else:
        logs = [
            log
            for log in file_system.glob(f"{base_path}/**")
            if Path(log).name == file_name
        ]

    return logs


def extract_records_with_pattern(pattern, log_path, log_file):

    file_system, log_path = infer_filesystem_type(log_path)

    logs = get_all_file_objects(log_path, log_file, file_system)

    keys = set()
    rows = []
    for log in logs:
        run_id = Path(log).parents[0].name
        experiment = (
            Path(log).parents[1].relative_to(log_path).as_posix().replace("/", ":")
        )
        keys.add("run_id")
        keys.add("experiment")
        metadata = parse_grok_multiline(log, pattern, file_system)

        row = {}
        row.update({"run_id": run_id})
        row.update({"experiment": experiment})
        for obj in metadata:
            keys.add(obj["name"])
            row.update({obj["name"]: obj["value"]})
        rows.append(row)

    logger.info(metadata)
    print_table(
        keys, rows, styles=STYLES, titles=TITLES, max_column_widths=MAX_COLUMN_WIDTHS
    )



"""Container API methods"""

import os
from pathlib import Path
import logging
from github import Github
from environs import Env
from mldock.platform_helpers import utils

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("mldock")
env = Env()
env.read_env()  # read .env file, if it exists


def check_available_templates(templates_root, **kwargs):
    """
    initialize mldock container project from template

    args:
        templates_root (str):
    kwargs:
        template_server (str): (default=local) template server to use
    """
    logger.info("Initializing Container Project")

    if kwargs.get("template_server", "local") == "local":
        template_dir = templates_root
        available_templates = [p.name for p in Path(template_dir).iterdir()]
    elif kwargs.get("template_server", "local") == "github":

        # set up in either ENV or configure local
        github_token = env("MLDOCK_GITHUB_TOKEN", default=None)
        github_org = env("MLDOCK_GITHUB_ORG", default="mldock")
        github_repo = env("MLDOCK_GITHUB_REPO", default="mldock-templates")
        github_branch = env("MLDOCK_GITHUB_REPO_BRANCH", default="main")

        github = Github(github_token)

        available_templates = utils.list_templates(
            github, ".", github_org, github_repo, github_branch, root=templates_root
        )

    return available_templates


def init_from_template(template_name, templates_root, src_directory, **kwargs):
    """
    initialize mldock container project from template

    args:
        template_name (str):
        templates_root (str):
        src_directory (str):
        container_only (bool): (default=False) only initialize the container directory
        template_server (str): (default=local) template server to use
    kwargs:
        container_only (bool): (default=False) only initialize the container directory
        template_server (str): (default=local) template server to use
        prediction_script (str): (default=None) optional prediction script to use
        trainer_script (str): (default=None) optional trainer script to use
    """
    logger.info("Initializing Container Project")

    if kwargs.get("template_server", "local") == "local":
        template_dir = templates_root

    elif kwargs.get("template_server", "local") == "github":

        # set up in either ENV or configure local
        github_token = env("MLDOCK_GITHUB_TOKEN", default=None)
        github_org = env("MLDOCK_GITHUB_ORG", default="mldock")
        github_repo = env("MLDOCK_GITHUB_REPO", default="mldock-templates")
        github_branch = env("MLDOCK_GITHUB_REPO_BRANCH", default="main")

        github = Github(github_token)

        template_dir = utils.download_from_git(
            github,
            template_name,
            github_org,
            github_repo,
            github_branch,
            root=templates_root,
        )

    logger.debug("Initializing template based on = {}".format(template_name))
    logger.debug("Template directory = {}".format(template_dir))

    logger.info("Setting up workspace")

    if kwargs.get("container_only", False):
        src_container_directory = os.path.join(src_directory, "container")
        utils._copy_boilerplate_to_dst(
            os.path.join(template_dir, template_name, "src/container"),
            src_container_directory,
            remove_first=True,
        )
    else:
        utils._copy_boilerplate_to_dst(
            os.path.join(template_dir, template_name, "src/"), src_directory
        )

    if kwargs.get("prediction_script", None) is not None:
        new_script_location = Path(src_directory, "prediction.py")
        logger.info(
            f"Copying {kwargs.get('prediction_script', None)} => {new_script_location}"
        )
        utils.copy_file(kwargs.get("prediction_script", None), new_script_location)

    if kwargs.get("trainer_script", None) is not None:
        new_script_location = Path(src_directory, "trainer.py")
        logger.info(
            f"Copying {kwargs.get('trainer_script', None)} => {new_script_location}"
        )
        utils.copy_file(kwargs.get("trainer_script", None), new_script_location)



from pathlib import Path
import tempfile
from pyarrow import fs
from mldock.platform_helpers.mldock.storage.pyarrow import upload_assets, download_assets

class TestPyarrow:

    # helpers
    @staticmethod
    def __create_textfile(my_path):
        """creates textfile and seeds it with msg"""
        # This currently assumes the path to the file exists.
        msg = 'test that this works'
        with open(my_path, 'w+') as file:
            file.write(msg)
        return msg

    # tests

    def test_local_upload_assets_successfully_upload_assets(self):
        """test local upload assets workflow"""
        local_file_system = fs.LocalFileSystem()

        with tempfile.TemporaryDirectory(suffix='data') as tmp_dir1:
            with tempfile.TemporaryDirectory() as tmp_dir2:
                txtfile = Path(tmp_dir1, 'data.txt').as_posix()
                _ = self.__create_textfile(txtfile)
                upload_assets(
                    file_system=local_file_system,
                    fs_base_path=tmp_dir2,
                    local_path=tmp_dir1,
                    storage_location='example',
                    zip_artifacts=False
                )

                files = [f.name for f in Path(tmp_dir2).glob("**/*") if f.is_file()]
                directories = [f.name for f in Path(tmp_dir2).glob("**/*") if f.is_dir()]

        
        assert "data.txt" in files, "Failure"
        assert "example" in directories, "Failure"

    def test_local_upload_assets_successfully_upload_assets_to_zip(self):
        """test local upload assets zip workflow"""
        local_file_system = fs.LocalFileSystem()

        with tempfile.TemporaryDirectory(suffix='data') as tmp_dir1:
            with tempfile.TemporaryDirectory() as tmp_dir2:
                txtfile = Path(tmp_dir1, 'data.txt').as_posix()
                _ = self.__create_textfile(txtfile)
                upload_assets(
                    file_system=local_file_system,
                    fs_base_path=tmp_dir2,
                    local_path=tmp_dir1,
                    storage_location='example',
                    zip_artifacts=True
                )

                files = [f.name for f in Path(tmp_dir2).glob("**/*") if f.is_file()]
                directories = [f.name for f in Path(tmp_dir2).glob("**/*") if f.is_dir()]

        
        assert "artifacts.zip" in files, "Failure"
        assert "example" in directories, "Failure"

    def test_local_download_assets_successfully_download_assets(self):
        """test local download assets workflow"""
        local_file_system = fs.LocalFileSystem()

        with tempfile.TemporaryDirectory(suffix='data') as tmp_dir1:
            with tempfile.TemporaryDirectory() as tmp_dir2:
                txtfile = Path(tmp_dir2, 'data.txt').as_posix()
                _ = self.__create_textfile(txtfile)
                download_assets(
                    file_system=local_file_system,
                    fs_base_path=tmp_dir2,
                    local_path=tmp_dir1,
                    storage_location='example'
                )

                files = [f.name for f in Path(tmp_dir2).glob("**/*") if f.is_file()]
                directories = [f.name for f in Path(tmp_dir2).glob("**/*") if f.is_dir()]

        
        assert "data.txt" in files, "Failure"
        assert "example" in directories, "Failure"





"""CONTAINER PROJECT MANAGEMENT COMMANDS"""
import os
import json
from pathlib import Path
import logging
import click

from mldock.config_managers.core import WorkingDirectoryManager
from mldock.config_managers.project import MLDockConfigManager
from mldock.platform_helpers import utils
from mldock.config_managers.cli import (
    PackageConfigManager,
    HyperparameterConfigManager,
    InputDataConfigManager,
    StageConfigManager,
    ModelConfigManager,
    EnvironmentConfigManager,
    CliConfigureManager,
)

from mldock.api.templates import init_from_template, check_available_templates

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")
MLDOCK_CONFIG_NAME = "mldock.json"


def reset_terminal():
    """clears the terminal view frame"""
    click.clear()


@click.group()
def project():
    """
    Commands to create, update and manage project projects and templates.
    """


@click.command()
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock project project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--no-prompt",
    is_flag=True,
    help="Do not prompt user, instead use the mldock config to initialize the project.",
)
@click.option(
    "--container-only", is_flag=True, help="Only inject new container assets."
)
@click.option(
    "--template",
    default=None,
    help="Directory containing mldock supported project template to use to initialize the new project.",
)
@click.option("--requirements", default=None, help="path to requirements file.")
@click.option(
    "--params",
    "-p",
    help="(Optional) Hyperparameter to be added in config.",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
)
@click.option(
    "--env_vars",
    "-e",
    help="(Optional) Environment Variables to be added in config.",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
)
@click.option(
    "--trainer-script",
    help="provide a path to your trainer script.",
    type=click.Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--prediction-script",
    help="provide a path to your prediction script.",
    type=click.Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.pass_obj
def init(obj, project_directory, **kwargs):
    """
    Command to initialize mldock enabled container project
    """
    no_prompt = kwargs.get("no_prompt", False)
    container_only = kwargs.get("container_only", False)
    template = kwargs.get("template", "generic")
    requirements = kwargs.get("requirements", None)
    params = kwargs.get("params", None)
    env_vars = kwargs.get("env_vars", None)
    trainer_script = kwargs.get("trainer_script", None)
    prediction_script = kwargs.get("prediction_script", None)

    reset_terminal()
    mldock_package_path = obj["mldock_package_path"]
    try:
        click.secho("Initializing MLDock project configuration", bg="blue", nl=True)

        if not Path(project_directory, "mldock.json").is_file():
            create_new = click.prompt(
                "No MLDOCK project found. Create?", type=bool, default="no"
            )
        else:
            create_new = False

        # check template config
        config_manager = CliConfigureManager()
        templates = config_manager.templates

        # get configured template server metadata
        # if not set, default to package default templates
        templates_root = templates.get(
            "templates_root", Path(mldock_package_path, "templates")
        )
        # if not set, default to local
        template_server = templates.get("server_type", "local")

        # get available templates
        available_templates = check_available_templates(
            templates_root=templates_root, template_server=template_server
        )

        mldock_manager = MLDockConfigManager(
            filepath=Path(project_directory, MLDOCK_CONFIG_NAME),
            create=create_new,
            available_templates=available_templates,
        )

        if not no_prompt:
            mldock_manager.setup_config()

        # get mldock config
        mldock_config = mldock_manager.get_config()

        if template is not None:
            mldock_manager.update_config(template=template)

        if params is not None:
            hyperparameters = mldock_config.get("hyperparameters", {})
            for param in params:
                key_, value_ = param
                hyperparameters.update({key_: value_})
            mldock_manager.update_config(hyperparameters=hyperparameters)

        if env_vars is not None:
            environment = mldock_config.get("environment", {})
            for env_var in env_vars:
                key_, value_ = env_var
                environment.update({key_: value_})
            mldock_manager.update_config(environment=environment)

        # finally delete original

        path_to_payload = Path(project_directory, "payload.json")
        if not path_to_payload.exists():
            path_to_payload.write_text(
                json.dumps({"feature1": 10, "feature2": "groupA"})
            )

        src_directory = os.path.join(
            project_directory, mldock_config.get("mldock_module_dir", "src")
        )

        _ = WorkingDirectoryManager(base_dir=project_directory)

        # create stages config
        stage_config_manager = StageConfigManager(
            config=mldock_config.get("stages", {})
        )
        # set input data channels
        input_data_channels = InputDataConfigManager(
            config=mldock_config.get("data", []),
            base_path=Path(project_directory, "data"),
        )

        # set model channels
        model_channels = ModelConfigManager(
            config=mldock_config.get("model", []),
            base_path=Path(project_directory, "model"),
        )

        # set hyperparameters
        hyperparameters = HyperparameterConfigManager(
            config=mldock_config.get("hyperparameters", {})
        )

        # set environment variables
        environment = EnvironmentConfigManager(
            config=mldock_config.get("environment", {})
        )

        if not no_prompt:
            stage_config_manager.ask_for_stages()
            # update stages
            mldock_manager.update_stages(stages=stage_config_manager.get_config())
            # update input data channels
            input_data_channels.ask_for_input_data_channels()
            input_data_channels.write_gitignore()
            mldock_manager.update_data_channels(data=input_data_channels.get_config())
            # update model channels
            model_channels.ask_for_model_channels()
            model_channels.write_gitignore()
            mldock_manager.update_model_channels(models=model_channels.get_config())
            # update hyperparameters
            hyperparameters.ask_for_hyperparameters()
            mldock_manager.update_hyperparameters(
                hyperparameters=hyperparameters.get_config()
            )
            # update environments
            environment.ask_for_env_vars()
            mldock_manager.update_env_vars(environment=environment.get_config())

        mldock_manager.write_file()

        # Get template specific files
        template = mldock_config["template"]

        init_from_template(
            template_name=template,
            templates_root=templates_root,
            src_directory=src_directory,
            container_only=container_only,
            template_server=template_server,
            trainer_script=trainer_script,
            prediction_script=prediction_script,
        )

        # always setup requirements in src/
        package_dir = mldock_manager.get_config().get(
            "requirements_dir", "src/requirements.txt"
        )
        requirement_file = Path(project_directory, package_dir)
        if requirements is not None:
            logger.info(f"Copying {requirements} => {requirement_file.as_posix()}")
            utils.copy_file(requirements, requirement_file)

        package_manager = PackageConfigManager(
            filepath=requirement_file, create=(requirement_file.is_file() == False)
        )
        package_manager.write_file()

        states = mldock_manager.get_state()

        for state in states:
            click.echo(click.style(state["name"], bg="blue"), nl=True)
            click.echo(click.style(state["message"], fg="white"), nl=True)

        logger.info("\nlocal project volume is ready! ヽ(´▽`)/")
    except Exception as exception:
        logger.error(exception)
        raise


@click.command()
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.pass_obj
def update(obj, project_directory):
    """
    Command to update mldock project.
    """
    reset_terminal()
    try:
        logger.info("Loading MLDock config")
        mldock_manager = MLDockConfigManager(
            filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
        )

        # get sagify_module_path name
        mldock_config = mldock_manager.get_config()

        _ = WorkingDirectoryManager(base_dir=project_directory)

        # create stages config
        stage_config_manager = StageConfigManager(
            config=mldock_config.get("stages", {})
        )
        # set input data channels
        input_data_channels = InputDataConfigManager(
            config=mldock_config.get("data", []),
            base_path=Path(project_directory, "data"),
        )
        # set model channels
        model_channels = ModelConfigManager(
            config=mldock_config.get("model", []),
            base_path=Path(project_directory, "model"),
        )

        # set hyperparameters
        hyperparameters = HyperparameterConfigManager(
            config=mldock_config.get("hyperparameters", {})
        )

        # set environment variables
        environment = EnvironmentConfigManager(
            config=mldock_config.get("environment", {})
        )

        stage_config_manager.ask_for_stages()
        # update stages
        mldock_manager.update_stages(stages=stage_config_manager.get_config())
        # update input data channels
        input_data_channels.ask_for_input_data_channels()
        input_data_channels.write_gitignore()
        mldock_manager.update_data_channels(data=input_data_channels.get_config())
        # update model channels
        model_channels.ask_for_model_channels()
        model_channels.write_gitignore()
        mldock_manager.update_model_channels(models=model_channels.get_config())
        # update hyperparameters
        hyperparameters.ask_for_hyperparameters()
        mldock_manager.update_hyperparameters(
            hyperparameters=hyperparameters.get_config()
        )
        # update environments
        environment.ask_for_env_vars()
        mldock_manager.update_env_vars(environment=environment.get_config())

        mldock_manager.write_file()

        click.clear()
        logger.info(obj["logo"])
        states = mldock_manager.get_state()

        for state in states:
            click.echo(click.style(state["name"], bg="blue"), nl=True)
            click.echo(click.style(state["message"], fg="white"), nl=True)

        logger.info("\nlocal project was updated! ヽ(´▽`)/")
    except Exception as exception:
        logger.error(exception)
        raise


@click.command()
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
def summary(project_directory):
    """
    Command to show summary for mldock container project
    """
    try:
        mldock_manager = MLDockConfigManager(
            filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
        )

        states = mldock_manager.get_state()

        for state in states:
            click.echo(click.style(state["name"], bg="blue"), nl=True)
            click.echo(click.style(state["message"], fg="white"), nl=True)

    except Exception as exception:
        logger.error(exception)
        raise


def add_commands(cli_group: click.group):
    """
    add commands to cli group
    args:
        cli (click.group)
    """
    cli_group.add_command(init)
    cli_group.add_command(update)
    cli_group.add_command(summary)


add_commands(project)



"""
    Container Project Config manager

    Config Managers that create and update the container project config.
    Currently supported:
        - mldock
"""
import os
import sys
import logging
from pathlib import Path
import click

from mldock.terminal import ChoiceWithNumbers, style_dropdown
from mldock.config_managers.core import BaseConfigManager
from mldock.platform_helpers.mldock import utils as mldock_utils
from mldock.platform_helpers import utils

logger = logging.getLogger("mldock")


class MLDockConfigManager(BaseConfigManager):
    """Hyperparameter Config Manager for mldock"""

    config = {
        "image_name": None,
        "template": "generic",
        "requirements_dir": "src/requirements.txt",
        "mldock_module_dir": "src",
        "container_dir": "container",
        "data": [],
        "environment": {},
        "hyperparameters": {},
        "model": [],
        "stages": {},
    }

    available_templates = None

    def __init__(self, filepath: str, create: bool = False, **kwargs):
        # dealing with weird os related path formats
        super().__init__()
        self.filepath = filepath
        self.available_templates = kwargs.get("available_templates", None)

        config = self.load_config(self.filepath, create=create)

        self.image_name_default = Path(filepath).parents[0].name

        self.config.update(config)

        if self.config.get("image_name", None) is None:
            # if is None override with default
            self.config["image_name"] = self.image_name_default

    def check_if_exists_else_create(self, file_name: str, create: bool):
        """check that mldock fileexists"""
        if not self.file_exists(file_name):
            if create:
                # deal with possiblity of nested directory
                utils._mkdir(Path(file_name).parents[0])
                # create file
                self.touch(file_name)
            else:
                logger.error(
                    (
                        "No MLDOCK container project found with "
                        "dir = '{CONTAINER_DIR}/'. Re-run `mldock project init` "
                        "and Confirm 'yes' to create.".format(
                            CONTAINER_DIR=Path(file_name).parents[0]
                        )
                    )
                )
                sys.exit(1)

    def setup_config(self):
        """run setup for configuration"""
        click.secho("Base Setup", bg="blue")
        self.ask_for_image_name()
        self.ask_for_template_name()
        self.ask_for_mldock_module_dir()
        self.ask_for_container_dir_name()
        self.ask_for_requirements_file_name()

    def update_config(self, **kwargs):
        """update configuration with given kwargs"""
        self.config.update(kwargs)

    def ask_for_image_name(self):
        """prompt user for image name"""
        image_name = click.prompt(
            text=click.style("Set your image name: ", fg="bright_blue"),
            default=self.config.get("image_name", self.image_name_default),
        )

        self.config.update({"image_name": image_name})

    def ask_for_template_name(self):
        """prompt user for container project template name"""

        click.secho("Set container project template name", bg="blue", nl=True)
        # would be awesome to fetch this from the template dir
        options = self.available_templates

        template_name = click.prompt(
            text=style_dropdown(
                group_name="template name",
                options=options,
                default=self.config.get("template", "generic"),
            ),
            type=ChoiceWithNumbers(options, case_sensitive=False),
            show_default=True,
            default=self.config.get("template", "generic"),
            show_choices=False,
        )

        self.config.update({"template": template_name})

    def ask_for_container_dir_name(self):
        """prompt user for container dir name"""
        container_dir_name = click.prompt(
            text=click.style("Set your container dir name: ", fg="bright_blue"),
            default=self.config.get("container_dir", "container"),
        )

        self.config.update({"container_dir": container_dir_name})

    def ask_for_mldock_module_dir(self):
        """prompt user for mldock module dir"""

        mldock_module_dir = click.prompt(
            text=click.style("Set mldock module dir: ", fg="bright_blue"),
            default=self.config.get("mldock_module_dir", "src"),
        )

        self.config.update({"mldock_module_dir": mldock_module_dir})

    @staticmethod
    def _format_data_node_item(item, base_path="data"):

        new_key = item["channel"]
        new_value = os.path.join(base_path, item["channel"], item["filename"])
        return "\t{KEY} : {VALUE}".format(KEY=new_key, VALUE=new_value)

    def _format_data_node(self):

        output = []

        for item in self.config["data"]:
            output.append(self._format_data_node_item(item, base_path="data"))

        return "\n".join(output)

    def _format_model_node(self):

        output = []

        for item in self.config["model"]:
            output.append(self._format_data_node_item(item, base_path="model"))

        return "\n".join(output)

    def _format_stage_node(self):

        output = []

        for key_, value_ in self.config["stages"].items():
            image_and_tag = "{}:{}".format(self.config["image_name"], value_["tag"])
            output.append("\t{} : {}".format(key_, image_and_tag))

        return "\n".join(output)

    @staticmethod
    def _format_nodes(config):

        output = []

        for key_, value_ in config.items():
            output.append("\t{} : {}".format(key_, value_))

        return "\n".join(output)

    def get_state(self):
        """pretty prints a json config to terminal"""
        config = self.config.copy()

        config.pop("stages")
        formatted_stages = self._format_stage_node()
        hyperparameters = config.pop("hyperparameters")
        config.pop("data")
        formatted_data_node = self._format_data_node()
        config.pop("model")
        formatted_model_node = self._format_model_node()
        environment = {}
        for key_, value_ in config.pop("environment").items():
            mldock_key = mldock_utils._format_key_as_mldock_env_var(
                key_, prefix="mldock"
            )
            environment.update({mldock_key: value_})

        states = []

        states.append({"name": "Base Setup", "message": self._format_nodes(config)})
        states.append({"name": "Data Channels", "message": formatted_data_node})
        states.append({"name": "Model Channels", "message": formatted_model_node})
        states.append({"name": "Stages", "message": formatted_stages})
        states.append(
            {"name": "Hyperparameters", "message": self._format_nodes(hyperparameters)}
        )
        states.append(
            {
                "name": "Environment Variables",
                "message": self._format_nodes(environment),
            }
        )

        return states

    def ask_for_requirements_file_name(self):
        """prompt user for image name"""
        requirements_file_name = click.prompt(
            text=click.style("Set full path to requirements: ", fg="bright_blue"),
            default=self.config.get("requirements_dir", "requirements.txt"),
        )

        self.config.update({"requirements_dir": requirements_file_name})

    def update_stages(self, stages: dict):
        """update stages node in .mldock config"""
        self.config.update({"stages": stages})

    def update_env_vars(self, environment: dict):
        """
        Update environment node in .mldock config

        Args:
            environment (dict): key,value dictionaries
        """
        self.config.update({"environment": environment})

    def update_hyperparameters(self, hyperparameters: dict):
        """
        Update hyperparameter node in .mldock config

        Args:
            hyperparameters (dict): key,value dictionaries
        """
        self.config.update({"hyperparameters": hyperparameters})

    def update_data_channels(self, data: dict):
        """
        Update data node in .mldock config

        Args:
            data (dict): key,value dictionaries
        """
        self.config.update({"data": data})

    def update_model_channels(self, models: dict):
        """
        Update data node in .mldock config

        Args:
            data (dict): key,value dictionaries
        """
        self.config.update({"model": models})



"""
    Core config managers

    Config managers that are reused across mldock cli tool.
"""
import sys
import os
import json
import logging
from pathlib import Path
import click

from mldock.platform_helpers import utils


logger = logging.getLogger("mldock")


class WorkingDirectoryManager:
    """Base config manager with basic read, write and update functionality."""

    def __init__(self, base_dir):
        self.base_dir = base_dir
        prompt_input = True

        if not self.file_exists(self.base_dir):
            prompt_input = click.prompt("File not found. Create?", default="yes")

            if prompt_input:
                self.base_dir.mkdir(parents=True, exist_ok=True)

        self.make_asset_dirs()

    def get_working_dir(self):
        """Get working directory path"""
        return self.base_dir

    def make_asset_dirs(self):
        """Create the directory structure, if not exists"""
        logger.debug(
            "Creating assets directories in working dir {} .".format(self.base_dir)
        )

        self.input_data_dir.mkdir(parents=False, exist_ok=True)
        self.input_config_dir.mkdir(parents=False, exist_ok=True)
        self.model_dir.mkdir(parents=False, exist_ok=True)
        self.output_data_dir.mkdir(parents=False, exist_ok=True)

    @staticmethod
    def file_exists(filename: str) -> bool:
        """Check if file exists

        Args:
            filename (str): path to file

        Returns:
            bool: whether file exists
        """
        return os.path.exists(filename)

    @property
    def input_data_dir(self):
        """Get input data directory path"""
        return Path(self.base_dir, "data")

    @property
    def input_config_dir(self):
        """Get input config directory path"""
        return Path(self.base_dir, "config")

    @property
    def model_dir(self):
        """Get model directory path"""
        return Path(self.base_dir, "model")

    @property
    def output_data_dir(self):
        """Get output data directory path"""
        return Path(self.base_dir, "output")


class BaseConfigManager:
    """Base config manager with basic read, write and update functionality."""

    config: dict = {}
    filepath: str = None

    @staticmethod
    def touch(path: str):
        """create a blank json file, seeded with {}

        Args:
            path (str): path to file
        """
        with open(path, "a") as file_:
            json.dump({}, file_)

    @staticmethod
    def file_exists(filename: str) -> bool:
        """Check if file exists

        Args:
            filename (str): path to file

        Returns:
            bool: whether file exists
        """
        return os.path.exists(filename)

    def check_if_exists_else_create(self, file_name: str, create: bool):
        """check that mldock fileexists"""
        if not self.file_exists(file_name):
            if create:
                # deal with possiblity of nested directory
                utils._mkdir(Path(file_name).parents[0])
                # create file
                self.touch(file_name)
            else:
                logger.error(
                    (
                        "file not found: '{DIRECTORY_PATH}/'. Please create.".format(
                            DIRECTORY_PATH=Path(file_name).parents[0]
                        )
                    )
                )
                sys.exit(1)

    def load_config(self, file_name: str, create: bool) -> dict:
        """loads config from file

        Args:
            filename (str): path to config to load
        Returns:
            dict: config
        """
        self.check_if_exists_else_create(file_name=file_name, create=create)

        with open(file_name, "r") as file_:
            config = json.load(file_)
            return config

    def pretty_print(self):
        """pretty prints a json config to terminal"""
        pretty_config = json.dumps(
            self.config, indent=4, separators=(",", ": "), sort_keys=True
        )
        logger.debug("{}\n".format(pretty_config))

    def write_file(self):
        """
        Trains ML model(s) locally
        :param dir: [str], source root directory
        :param docker_tag: [str], the Docker tag for the image
        :param image_name: [str], The name of the Docker image
        """
        with open(self.filepath, "w") as config_file:
            json.dump(self.config, config_file, indent=4, sort_keys=True)

    def get_config(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config



"""LOGGING-BASED MLDOCK TRACKER COMPONENTS"""
import atexit
import io
import uuid
import warnings
import logging
from pathlib import Path
import tempfile
from pyarrow import fs

from mldock.platform_helpers import utils

logger = logging.getLogger("mldock")
logger.setLevel(logging.INFO)


class ExperimentTracker:
    """
    Experiment Tracker that supports logging metrics
    and params. Finally, uploading to remote on exit if fs_base_path is provided
    """

    # pylint: disable=too-many-instance-attributes
    metrics: dict = None
    params: dict = None

    def __init__(self, experiment_name, experiment_logger, file_system, **kwargs):

        self.file_system = file_system
        self.fs_base_path = kwargs.get("fs_base_path", None)
        self.base_path = kwargs.get("base_path", ".")
        self.manifest = kwargs.get("manifest", {})

        self.experiment_name = experiment_name
        self.run_id = str(uuid.uuid4())

        self.experiment_logger = experiment_logger
        # use a certain type of StreamHandler
        self.buffer = io.StringIO()
        stream_handler = logging.StreamHandler(self.buffer)

        # pytest: disable=consider-using-with
        self.tmp_dir = tempfile.TemporaryDirectory(suffix=None, prefix=None, dir=None)
        self.artifact_dir = Path(
            self.tmp_dir.name, self.base_path, self.experiment_name, self.run_id
        )
        Path(self.artifact_dir).mkdir(parents=True, exist_ok=True)

        file_handler = logging.FileHandler(
            Path(self.artifact_dir, "logs.txt"), mode="a"
        )

        # add it to experiment_logger
        self.experiment_logger.addHandler(stream_handler)
        self.experiment_logger.addHandler(file_handler)

        self.metrics = {}
        self.params = {}

        # register write_artifacts to fire on script exit.
        atexit.register(self.write_artifacts)

    def log(self, msg, *args, **kwargs):
        """log message to log stream"""

        level = kwargs.pop("level", logging.INFO)
        self.experiment_logger.log(level, msg, *args, **kwargs)

    def log_metric(self, name, value):
        """log a metric to log stream"""
        self.metrics.update({name: value})
        msg = "metric: {NAME}={VALUE};".format(NAME=name, VALUE=value)
        self.log(msg)

    def log_metrics(self, metrics: dict):
        """log metrics to log stream"""
        self.metrics.update(metrics)
        for name, value in metrics.items():
            self.log_metric(name, value)

    def log_param(self, name, value):
        """log a param to log stream"""
        self.params.update({name: value})
        msg = "param: {NAME}={VALUE};".format(NAME=name, VALUE=value)
        self.log(msg)

    def log_params(self, params: dict):
        """log params to log stream"""
        self.params.update(params)
        for name, value in params.items():
            self.log_param(name, value)

    def upload_assets(self, fs_base_path):
        """Uploads logs to specified file-system"""
        # strip the scheme if it is provided
        fs_base_path = utils.strip_scheme(fs_base_path)

        # create full artifacts base path
        artifacts_base_path = Path(
            fs_base_path, self.base_path, self.experiment_name, self.run_id
        )

        local = fs.LocalFileSystem()

        file_selector = fs.FileSelector(self.artifact_dir.as_posix(), recursive=True)
        for file in local.get_file_info(file_selector):
            src_path = Path(file.path)
            file_name = src_path.name
            dst_path = Path(artifacts_base_path, file_name)

            if isinstance(self.file_system, fs.LocalFileSystem):
                artifacts_base_path.mkdir(parents=True, exist_ok=True)
                self.file_system.copy_file(src_path.as_posix(), dst_path.as_posix())
            else:
                self.file_system.upload(src_path.as_posix(), dst_path.as_posix())

    def write_metadata(self):
        """write metadata file to experiment tmp location"""
        self.manifest.update({"params": self.params, "metrics": self.metrics})
        utils._write_json(
            obj=self.manifest, file_path=Path(self.artifact_dir, "manifest.json")
        )

    def write_artifacts(self):
        """Write experiment artifacts"""
        try:
            if self.fs_base_path is None:
                warnings.warn(
                    "Skipping syncronization of run = '{RUN}' for experiment = '{EXPERIMENT}'. "
                    "To syncronize set 'fs_base_path' to "
                    "either local filepath, s3 or gs.".format(
                        EXPERIMENT=self.experiment_name, RUN=self.run_id
                    )
                )
            else:
                # writes the final metadata file
                self.write_metadata()

                # upload assets to file-system
                self.upload_assets(fs_base_path=self.fs_base_path)
        except Exception as exception:
            raise "Failed to write artifacts" from exception
        finally:
            self.tmp_dir.cleanup()



"""LOGS PARAMS COMMANDS"""
import logging
import click
from mldock.api.logs import extract_records_with_pattern

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")
MLDOCK_CONFIG_NAME = "mldock.json"


@click.group()
def params():
    """
    Commands to manage and interact with params logs.
    """


@click.command()
@click.option("--log-path", type=str, help="a grok pattern")
@click.option("--log-file", type=str, default="logs.txt", help="file name")
def show(log_path, log_file):
    """show params for all runs as a table"""

    pattern = r"param: %{GREEDYDATA:name}=%{NUMBER:value};"

    extract_records_with_pattern(pattern, log_path, log_file)


params.add_command(show)



"""TEMPLATES COMMANDS"""
import logging
from pathlib import Path
import click

from mldock.platform_helpers import utils

click.disable_unicode_literals_warning = True
logger = logging.getLogger("mldock")
MLDOCK_CONFIG_NAME = "mldock.json"


@click.group()
def templates():
    """
    Commands to create, update and manage container projects and templates.
    """


@click.command()
@click.option(
    "--name",
    help="Container name, used to name directories,etc.",
    required=True,
    type=str,
)
@click.option(
    "--project_directory",
    "--dir",
    "-d",
    help="mldock container project.",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None,
    ),
)
@click.option(
    "--out",
    help="Destination of template should be stored once created.",
    required=True,
    type=str,
)
def create(name, project_directory, out):
    """
    Command to create a mldock enabled container template
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception(
                (
                    "Path '{}' was not an mldock project. "
                    "Confirm this directory is correct, otherwise "
                    "create one.".format(project_directory)
                )
            )

        mldock_src_path = Path(project_directory, "src")

        destination_path = Path(out, name)
        destination_path.mkdir(parents=False, exist_ok=True)

        destination_src_path = Path(destination_path, "src")

        utils._copy_boilerplate_to_dst(mldock_src_path, destination_src_path)
    except Exception as exception:
        logger.error(exception)
        raise


templates.add_command(create)
